#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for SchoolTool timetable wizard.
"""

import unittest
import doctest
import datetime
from pprint import pprint

from zope.component import provideAdapter
from zope.interface import Interface
from zope.i18n import translate
from zope.location.location import locate
from zope.publisher.browser import TestRequest
from zope.schema import TextLine

from schooltool.app.browser import testing as schooltool_setup
from schooltool.app.app import SimpleNameChooser
from schooltool.app.app import getApplicationPreferences
from schooltool.app.app import getSchoolToolApplication
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.common import format_time_range
from schooltool.testing import setup
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.interfaces import IHaveTimetables
from schooltool.timetable.daytemplates import TimeSlot
from schooltool.timetable.schedule import Period
from schooltool.timetable.timetable import Timetable
from schooltool.timetable.timetable import TimetableContainer
from schooltool.timetable.browser.ttwizard import TimetableWizard
from schooltool.testing.util import format_table


class SchoolYearStub(object):
    def __init__(self, first, last):
        self.first = first
        self.last = last


def setUp(test):
    """Test setup.

    Sets up enough of Zope 3 to be able to render page templates.

    Creates a SchoolTool application and makes it both the current site
    and the containment root.  The application object is available as
    a global named `app` in all doctests.
    """
    schooltool_setup.setUp(test)
    setup.setUpSessions()
    setup.setUpApplicationPreferences()
    provideAdapter(SimpleNameChooser, (ITimetableContainer,))
    provideAdapter(getApplicationPreferences,
                   (ISchoolToolApplication,), IApplicationPreferences)
    provideAdapter(getSchoolToolApplication, (None,), ISchoolToolApplication)


    app = test.globs['app'] = setup.setUpSchoolToolSite()

    schoolyear = app['schoolyear'] = SchoolYearStub(
        datetime.date(2010, 07, 01),
        datetime.date(2011, 06, 15))
    locate(schoolyear, app, 'schoolyear')
    timetables =test.globs['timetables'] = TimetableContainer()
    locate(timetables, app['schoolyear'], 'timetables')

    provideAdapter(lambda container: schoolyear,
                   (ITimetableContainer, ),
                   IHaveTimetables)


def tearDown(test):
    """Test cleanup."""
    schooltool_setup.tearDown(test)


def print_day_templates(templates, filter=None):
    cols = []
    for day in templates.values():
        col = [day.title]
        for item in day.values():
            if (filter is not None and
                not filter(item)):
                col.append('')
            elif isinstance(item, Period):
                cell = item.title or ''
                if item.activity_type is not None:
                    cell = '%s %s' % (item.activity_type, cell)
                col.append(cell or '?')
            elif isinstance(item, TimeSlot):
                cell = format_time_range(item.tstart, item.duration)
                if item.activity_type is not None:
                    cell = '%s %s' % (item.activity_type, cell)
                col.append(cell or '?')
            else:
                col.append(item.__class__.__name_)
        cols.append(col)
    max_len = max([len(c) for c in cols])
    cols = [c + ['']*(max_len-len(c)) for c in cols]
    rows = map(None, *cols)
    print format_table(rows, header_rows=1)


def print_timetable(timetable):
    print "Timetable '%s'" % timetable.title
    print 'Periods (%s)' % timetable.periods.__class__.__name__
    print_day_templates(timetable.periods.templates)
    print 'Time slots (%s)' % timetable.time_slots.__class__.__name__
    print_day_templates(timetable.time_slots.templates)


def doctest_wizard_getSessionData():
    """Unit test for getSessionData.

    This function is used as a method for both Step and TimetableWizard
    classes (and subclasses of the former).

        >>> from schooltool.timetable.browser.ttwizard import Step
        >>> context = timetables
        >>> request = TestRequest()
        >>> step = Step(context, request)
        >>> data = step.getSessionData()
        >>> data
        <...SessionPkgData...>
        >>> data['something'] = 42

        >>> request = TestRequest()
        >>> step = Step(context, request)
        >>> data['something']
        42

        >>> view = TimetableWizard(context, request)
        >>> data['something']
        42

    """


def doctest_wizard_ChoiceStep():
    """Unit test for ChoiceStep

        >>> from schooltool.timetable.browser.ttwizard import ChoiceStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = ChoiceStep(context, request)

    ChoiceStep wants some attributes

        >>> view.key = 'meaning'
        >>> view.question = 'What is the meaning of life?'
        >>> view.choices = [('42', "Fourty two"),
        ...                 ('huh', "I don't know")]

        >>> print view()
        <BLANKLINE>
        ...What is the meaning of life?...
        ...<input class="button-ok" type="submit" name="NEXT.0"
                  value="Fourty two" />
        ...<input class="button-ok" type="submit" name="NEXT.1"
                  value="I don't know" />
        ...

    Update does something if you choose a valid choice

        >>> view.update()
        False

        >>> view.request = TestRequest(form={'NEXT.0': "Whatever"})
        >>> view.update()
        True
        >>> view.getSessionData()[view.key]
        '42'

        >>> view.request = TestRequest(form={'NEXT.1': "Whatever"})
        >>> view.update()
        True
        >>> view.getSessionData()[view.key]
        'huh'
    """


def doctest_wizard_FormStep():
    """Unit test for FormStep

    FormStep needs to be subclassed, and a subclass has to provide a `schema`
    attribute

        >>> from schooltool.timetable.browser.ttwizard import FormStep
        >>> class SampleFormStep(FormStep):
        ...     description = "Informative text shown above the form."
        ...     class schema(Interface):
        ...         a_field = TextLine(title=u"A field")
        ...         b_field = TextLine(title=u"B field")

    The constructor sets up input widgets.

        >>> context = timetables
        >>> request = TestRequest()
        >>> view = SampleFormStep(context, request)
        >>> view.a_field_widget
        <...TextWidget...>
        >>> view.b_field_widget
        <...TextWidget...>

    The `widgets` method returns all widgets in oroder

        >>> view.widgets() == [view.a_field_widget, view.b_field_widget]
        True

    Calling the view renders the form

        >>> print view()
        <BLANKLINE>
        ...
        <form class="plain" method="POST" action="http://127.0.0.1">
          <p>Informative text shown above the form.</p>
              <div class="row">
                  <div class="label">
                    <label for="field.a_field" title="">A field</label>
                  </div>
                  <div class="field"><input class="textType" id="field.a_field" name="field.a_field" size="20" type="text" value=""  /></div>
              </div>
              <div class="row">
                  <div class="label">
                    <label for="field.b_field" title="">B field</label>
                  </div>
                  <div class="field"><input class="textType" id="field.b_field" name="field.b_field" size="20" type="text" value=""  /></div>
              </div>
          <div class="controls">
            <input type="submit" class="button-ok" name="NEXT"
                   value="Next" />
            <input type="submit" class="button-cancel" name="CANCEL"
                   value="Cancel" />
          </div>
        </form>
        ...

    An error message may come from the `error` attribute.

        >>> view.error = "A terrible error has occurred!"
        >>> print view()
        <BLANKLINE>
        ...
        <form class="plain" method="POST" action="http://127.0.0.1">
          <p>Informative text shown above the form.</p>
          <div class="error">A terrible error has occurred!</div>
        ...

    """


def doctest_wizard_FirstStep():
    """Unit test for FirstStep

        >>> from schooltool.timetable.browser.ttwizard import FirstStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = FirstStep(context, request)

        >>> print view()
        <BLANKLINE>
        ...<input class="textType" id="field.title" name="field.title"
                  size="20" type="text" value="default" />...
        ...<input type="submit" class="button-ok" name="NEXT"
                  value="Next" />
        ...

    FirstStep.update can take the title from the request and put it into
    the session.

        >>> request = TestRequest(form={'field.title': u'Sample Timetable'})
        >>> view = FirstStep(context, request)
        >>> view.update()
        True

        >>> view.getSessionData()['title']
        u'Sample Timetable'

    If the form is incomplete, update says so by returning False

        >>> request = TestRequest(form={'field.title': u''})
        >>> view = FirstStep(context, request)
        >>> view.update()
        False

    The next step is always CycleStep

        >>> view.next()
        <...CycleStep...>

    """


def doctest_wizard_CycleStep():
    """Unit test for CycleStep

        >>> from schooltool.timetable.browser.ttwizard import CycleStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = CycleStep(context, request)
        >>> session = view.getSessionData()

    The next step is DayEntryStep for rotating cycle

        >>> session['cycle'] = 'rotating'
        >>> view.next()
        <...DayEntryStep...>

    The next step is IndependentDaysStep for weekly cycle

        >>> session['cycle'] = 'weekly'
        >>> view.next()
        <...IndependentDaysStep...>

    The session variable day_names is filled with weekday
    names in view.update() if and only if the cycle is weekly

        >>> view.request.form['NEXT.1'] = 'Rotating' # rotating
        >>> view.update()
        True
        >>> 'day_names' in session
        False

        >>> view.request.form['NEXT.0'] = 'weekly' # weekly
        >>> view.update()
        True
        >>> session['day_names']
        [u'Monday', u'Tuesday', u'Wednesday', u'Thursday', u'Friday',
         u'Saturday', u'Sunday']

    """


def doctest_wizard_DayEntryStep():
    r"""Unit test for DayEntryStep

        >>> from schooltool.timetable.browser.ttwizard import DayEntryStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = DayEntryStep(context, request)

        >>> print view()
        <BLANKLINE>
        ...<textarea cols="60" id="field.days" name="field.days"
                     rows="15" ></textarea>...

    DayEntryStep.update wants at least one day name

        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least one day name.

        >>> request = TestRequest(form={'field.days': u''})
        >>> view = DayEntryStep(context, request)
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least one day name.

        >>> request = TestRequest(form={'field.days': u'\n\n\n'})
        >>> view = DayEntryStep(context, request)
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least one day name.

        >>> request = TestRequest(form={'field.days': u'A\nB\nA\n'})
        >>> view = DayEntryStep(context, request)
        >>> view.update()
        False
        >>> print translate(view.error)
        Please make sure the day names are unique.

        >>> request = TestRequest(form={'field.days': u'A\nB\n\n'})
        >>> view = DayEntryStep(context, request)
        >>> view.update()
        True
        >>> view.error

        >>> view.getSessionData()['day_names']
        [u'A', u'B']

    The next page is IndependentDaysStep.

        >>> view.next()
        <...IndependentDaysStep...>

    """


def doctest_wizard_IndependentDaysStep():
    r"""Unit test for IndependentDaysStep

        >>> from schooltool.timetable.browser.ttwizard \
        ...                                         import IndependentDaysStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = IndependentDaysStep(context, request)

    The next step is SimpleSlotEntryStep if days are similar:

        >>> view.getSessionData()['similar_days'] = True
        >>> view.next()
        <...ttwizard.SimpleSlotEntryStep...>

    The next step cares about day names in the session.

        >>> view.getSessionData()['day_names'] = ['Day A', 'Day B']

    The next step is WeeklySlotEntryStep if each day is different and you chose
    the weekly cycle:

        >>> view.getSessionData()['similar_days'] = False
        >>> view.getSessionData()['cycle'] = 'weekly'
        >>> view.next()
        <...ttwizard.WeeklySlotEntryStep...>

    If you chose the rotating cycle, you will be asked another question
    about the time model:

        >>> view.getSessionData()['cycle'] = 'rotating'
        >>> view.next()
        <...ttwizard.SequentialModelStep...>

    """


def doctest_wizard_SequentialModelStep():
    r"""Unit test for SequentialModelStep

        >>> from schooltool.timetable.browser.ttwizard import SequentialModelStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = SequentialModelStep(context, request)

    If you choose the weekly cycle, you will be redirected to
    WeeklySlotEntryStep:

        >>> session = view.getSessionData()
        >>> session['time_model'] = 'weekly'
        >>> view.next()
        <...ttwizard.WeeklySlotEntryStep...>

    Otherwise, the rotating cycle will be used:

        >>> session['time_model'] = 'rotating'
        >>> view.next()
        <...ttwizard.RotatingSlotEntryStep...>

    """


def doctest_wizard_SimpleSlotEntryStep():
    r"""Unit test for SimpleSlotEntryStep

        >>> from schooltool.timetable.browser.ttwizard import SimpleSlotEntryStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = SimpleSlotEntryStep(context, request)

        >>> print view()
        <BLANKLINE>
        ...<textarea cols="60" id="field.times" name="field.times"
                     rows="15" >...</textarea>...

    SimpleSlotEntryStep.update wants at least one slot

        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least one time slot.

        >>> request = TestRequest(form={'field.times': u''})
        >>> view = SimpleSlotEntryStep(context, request)
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least one time slot.

        >>> request = TestRequest(form={'field.times': u'\n\n\n'})
        >>> view = SimpleSlotEntryStep(context, request)
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least one time slot.

        >>> request = TestRequest(form={'field.times': u'not a time\n\n\n'})
        >>> view = SimpleSlotEntryStep(context, request)
        >>> view.update()
        False
        >>> print translate(view.error)
        Not a valid time slot: not a time.

        >>> request = TestRequest(form={'field.times': u'unicode is tr\u00efcky'})
        >>> view = SimpleSlotEntryStep(context, request)
        >>> view.update()
        False
        >>> translate(view.error)
        u'Not a valid time slot: unicode is tr\xefcky.'

    Let's cover the successful case.  We will need day names in the session.

        >>> request = TestRequest(form={'field.times': u'9:30-10:25\n\n'})
        >>> view = SimpleSlotEntryStep(context, request)
        >>> view.getSessionData()['day_names'] = ['DA', 'DB', 'DC']
        >>> view.update()
        True
        >>> view.error

    The time_slots structure will be filled with as many duplicates of the
    slot times as there are day names.

        >>> pprint(view.getSessionData()['time_slots'])
        [[(datetime.time(9, 30), datetime.timedelta(0, 3300))],
         [(datetime.time(9, 30), datetime.timedelta(0, 3300))],
         [(datetime.time(9, 30), datetime.timedelta(0, 3300))]]

    The text area contains one slot per line; extra spaces are stripped;
    empty lines are ignored.

    The next page is the named periods page.

        >>> view.next()
        <...ttwizard.NamedPeriodsStep...>

    """


def doctest_wizard_RotatingSlotEntryStep():
    r"""Unit test for RotatingSlotEntryStep

        >>> from schooltool.timetable.browser.ttwizard import RotatingSlotEntryStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = RotatingSlotEntryStep(context, request)

        >>> view.getSessionData()['day_names'] = ['Oneday', 'Twoday']

        >>> view.dayNames()
        ['Oneday', 'Twoday']

    At first we get a table with one empty row of input fields:

        >>> print view()
        <BLANKLINE>
        ...
            <tr>
              <th>Oneday</th>
              <th>Twoday</th>
            </tr>
            <tr>
              <td>
                <textarea rows="12" cols="15" name="times.0">8:00 - 8:45
        9:05 - 9:50
        </textarea>
              </td>
              <td>
                <textarea rows="12" cols="15" name="times.1">8:00 - 8:45
        9:05 - 9:50
        </textarea>
              </td>
            </tr>
        </table>
        ...

        >>> print view.error
        None

    SlotEntryStep.update accepts days without any slots (so you could
    have a day without lessons in the middle of the cycle, or have
    lessons on Saturday, but not Sunday):

        >>> view.request.form['times.0'] = u'9:30 - 10:25'
        >>> view.request.form['times.1'] = u''
        >>> view.update()
        True

        >>> print view()
        <BLANKLINE>
        ...
        <tr>
          <th>Oneday</th>
          <th>Twoday</th>
        </tr>
        <tr>
          <td>
            <textarea rows="12" cols="15" name="times.0">9:30 - 10:25</textarea>
          </td>
          <td>
            <textarea rows="12" cols="15" name="times.1"></textarea>
          </td>
        </tr>
        </table>
        ...

    If we provide an invalid interval, an error message will be shown:

        >>> view.request.form['times.1'] = u'9:15 - 10:73'
        >>> view.update()
        False
        >>> print translate(view.error)
        Not a valid time slot: 9:15 - 10:73.

    If we provide both fields, the form will be parsed successfully:

        >>> view.request.form['times.1'] = u'9:15 - 10:10\n10:35 - 11:20'
        >>> view.update()
        True

        >>> pprint(view.getSessionData()['time_slots'])
        [[(datetime.time(9, 30), datetime.timedelta(0, 3300))],
         [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
          (datetime.time(10, 35), datetime.timedelta(0, 2700))]]

    The next page is the period naming step.

        >>> view.next()
        <...ttwizard.NamedPeriodsStep...>

    """


def doctest_wizard_WeeklySlotEntryStep():
    r"""Unit test for WeeklySlotEntryStep

        >>> from schooltool.timetable.browser.ttwizard import WeeklySlotEntryStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = WeeklySlotEntryStep(context, request)
        >>> view.getSessionData()['cycle'] = 'weekly'

        >>> view.dayNames()
        [u'Monday', u'Tuesday', u'Wednesday', u'Thursday', u'Friday',
         u'Saturday', u'Sunday']

    At first we get a table with one empty row of input fields:

        >>> print view()
        <BLANKLINE>
        ...
        <tr>
          <th>Monday</th>
          ...
          <th>Sunday</th>
        </tr>
        <tr>
          <td>
            <textarea rows="12" cols="15" name="times.0">8:00 - 8:45
        9:05 - 9:50
        </textarea>
          </td>
          ...
          <td>
            <textarea rows="12" cols="15" name="times.4">8:00 - 8:45
        9:05 - 9:50
        </textarea>
          </td>
          <td>
            <textarea rows="12" cols="15" name="times.5"></textarea>
          </td>
          <td>
            <textarea rows="12" cols="15" name="times.6"></textarea>
          </td>
        </tr>
        </table>
        ...

        >>> print view.error
        None

    SlotEntryStep.update accepts days without any slots:

        >>> view.request.form['times.0'] = u'9:30 - 10:25'
        >>> view.request.form['times.5'] = u''
        >>> view.update()
        True

        >>> print view()
        <BLANKLINE>
        ...
        <tr>
          <td>
            <textarea rows="12" cols="15" name="times.0">9:30 - 10:25</textarea>
          </td>
          ...
          <td>
            <textarea rows="12" cols="15" name="times.5"></textarea>
          </td>
          ...
        </tr>
        </table>
        ...

    If we provide an invalid interval, an error message will be shown:

        >>> view.request.form['times.1'] = u'9:15 - 10:73'
        >>> view.update()
        False
        >>> print translate(view.error)
        Not a valid time slot: 9:15 - 10:73.

    If we provide both fields, the form will be parsed successfully:

        >>> for i in range(1, 5):
        ...     field_name = 'times.%d' % i
        ...     view.request.form[field_name] = u'9:15 - 10:10\n10:35 - 11:20'
        >>> view.update()
        True

        >>> pprint(view.getSessionData()['time_slots'])
        [[(datetime.time(9, 30), datetime.timedelta(0, 3300))],
         [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
          (datetime.time(10, 35), datetime.timedelta(0, 2700))],
         [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
          (datetime.time(10, 35), datetime.timedelta(0, 2700))],
         [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
          (datetime.time(10, 35), datetime.timedelta(0, 2700))],
         [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
          (datetime.time(10, 35), datetime.timedelta(0, 2700))],
         [],
         []]

    The next page is the period naming step.

        >>> view.next()
        <...ttwizard.NamedPeriodsStep...>

    """


def doctest_wizard_NamedPeriodsStep():
    r"""Unit test for NamedPeriodsStep

        >>> from schooltool.timetable.browser.ttwizard \
        ...     import NamedPeriodsStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = NamedPeriodsStep(context, request)

    The next step is PeriodNamesStep if perods should be named:

        >>> view.getSessionData()['named_periods'] = True
        >>> view.next()
        <class 'schooltool.timetable.browser.ttwizard.PeriodNamesStep'>

    If periods are not named, go straight to the homeroom step, and compute
    the default period names:

        >>> from datetime import time, timedelta
        >>> periods = [(time(9, 0), timedelta(minutes=50)),
        ...            (time(12, 35), timedelta(minutes=50)),
        ...            (time(14, 15), timedelta(minutes=55))]
        >>> periods2 = [(time(9, 10), timedelta(minutes=50)),
        ...             (time(12, 35), timedelta(minutes=50))]
        >>> view.getSessionData()['time_slots'] = [periods] * 2 + [periods2]
        >>> view.getSessionData()['named_periods'] = False

        >>> view.next()
        <...ttwizard.HomeroomStep...>

        >>> pprint(view.getSessionData()['periods_order'])
        [['09:00-09:50', '12:35-13:25', '14:15-15:10'],
         ['09:00-09:50', '12:35-13:25', '14:15-15:10'],
         ['09:10-10:00', '12:35-13:25']]

    """


def doctest_wizard_PeriodNamesStep():
    r"""Unit test for PeriodNamesStep

        >>> from schooltool.timetable.browser.ttwizard import PeriodNamesStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = PeriodNamesStep(context, request)

    The number of the required periods depends on the maximum number
    of slots:

        >>> session = view.getSessionData()
        >>> interval = (datetime.time(9, 15), datetime.timedelta(0, 3300))
        >>> view.getSessionData()['time_slots'] = \
        ...     [[interval] * 3, [interval] * 5, [interval] * 2]
        >>> view.requiredPeriods()
        5
        >>> del session['time_slots']

    The view asks the user to enter at least 3 period names:

        >>> default_slots = [[interval] * 3]
        >>> view.getSessionData()['time_slots'] = default_slots
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least 3 periods.

        >>> request = TestRequest(form={'field.periods': u''})
        >>> view = PeriodNamesStep(context, request)
        >>> view.getSessionData()['time_slots'] = default_slots
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least 3 periods.

        >>> request = TestRequest(form={'field.periods': u'\n\n\n'})
        >>> view = PeriodNamesStep(context, request)
        >>> view.getSessionData()['time_slots'] = default_slots
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least 3 periods.

        >>> request = TestRequest(form={'field.periods': u'A\nB\nB\n'})
        >>> view = PeriodNamesStep(context, request)
        >>> view.getSessionData()['time_slots'] = default_slots
        >>> view.update()
        False
        >>> print translate(view.error)
        Please make sure the period names are unique.

        >>> request = TestRequest(form={'field.periods': u'A\n'})
        >>> view = PeriodNamesStep(context, request)
        >>> view.getSessionData()['time_slots'] = default_slots
        >>> view.update()
        False
        >>> print translate(view.error)
        Please enter at least 3 periods.

        >>> request = TestRequest(form={'field.periods': u'A\nB\nC\nD'})
        >>> view = PeriodNamesStep(context, request)
        >>> view.getSessionData()['time_slots'] = default_slots
        >>> view.update()
        True
        >>> view.error

        >>> view.getSessionData()['period_names']
        [u'A', u'B', u'C', u'D']

    The next page is HomeroomStep.

        >>> view.next()
        <class 'schooltool.timetable.browser.ttwizard.PeriodSequenceSameStep'>

    """


def doctest_wizard_PeriodSequenceSameStep():
    r"""Unit test for PeriodSequenceSameStep

        >>> from schooltool.timetable.browser.ttwizard \
        ...     import PeriodSequenceSameStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = PeriodSequenceSameStep(context, request)

    The next step is PeriodOrderSimple if periods are the same:

        >>> view.getSessionData()['periods_same'] = True
        >>> view.next()
        <class 'schooltool.timetable.browser.ttwizard.PeriodOrderSimple'>

    The next step is PeriodOrderComplex if periods are not the same:

        >>> view.getSessionData()['periods_same'] = False
        >>> view.next()
        <class 'schooltool.timetable.browser.ttwizard.PeriodOrderComplex'>

    """


def doctest_wizard_PeriodOrderSimple():
    """Unit test for PeriodOrderSimple view

        >>> from schooltool.timetable.browser.ttwizard import PeriodOrderSimple
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = PeriodOrderSimple(context, request)

    Let's say we have some periods:

        >>> view.getSessionData()['period_names'] = ['A', 'B', 'C', 'D']

    Our view lets the template easily access them:

        >>> view.periods()
        ['A', 'B', 'C', 'D']

    The number of period dropdowns is the maximum of slots in a day:

        >>> time_slots = [
        ...     [(datetime.time(9, 30), datetime.timedelta(0, 3300))],
        ...     [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
        ...      (datetime.time(10, 35), datetime.timedelta(0, 2700)),
        ...      (datetime.time(10, 35), datetime.timedelta(0, 2700)),]]
        >>> view.getSessionData()['time_slots'] = time_slots
        >>> view.numPeriods()
        3

    If we render the view, we get a list of dropdowns with consecutive
    periods selected:

        >>> print view()
        <BLANKLINE>
        ...
          <div class="row">
            <select name="period_0">
              <option selected="selected">A</option>
              <option>B</option>
              <option>C</option>
              <option>D</option>
            </select>
          </div>
          <div class="row">
            <select name="period_1">
              <option>A</option>
              <option selected="selected">B</option>
              <option>C</option>
              <option>D</option>
            </select>
          </div>
          <div class="row">
            <select name="period_2">
              <option>A</option>
              <option>B</option>
              <option selected="selected">C</option>
              <option>D</option>
            </select>
          </div>
        ...

    When the user shuffles the dropdowns and submits them, the order
    of periods is changed:

        >>> request = TestRequest(form={'period_0': 'D', 'period_1': 'C',
        ...                             'period_2': 'B', 'period_3': 'A'})
        >>> view = PeriodOrderSimple(context, request)
        >>> view.getSessionData()['period_names'] = ['A', 'B', 'C', 'D',
        ...                                          'E', 'F']
        >>> view.getSessionData()['day_names'] = ['1', '2']
        >>> view.getSessionData()['time_slots'] = time_slots

        >>> view.update()
        True
        >>> print view.getSessionData()['periods_order']
        [['D', 'C', 'B'], ['D', 'C', 'B']]

    The next step is always the homeroom step:

        >>> view.next()
        <class 'schooltool.timetable.browser.ttwizard.HomeroomStep'>

    If not all periods are in the request, update fails and the user
    gets an error.  This is unlikely in real life as the dropdowns
    have values preselected:

        >>> request = TestRequest(form={'period_0': 'D', 'period_1': 'C'})
        >>> view = PeriodOrderSimple(context, request)
        >>> view.getSessionData()['period_names'] = ['A', 'B', 'C', 'D']
        >>> view.getSessionData()['time_slots'] = time_slots
        >>> view.update()
        False
        >>> print translate(view.error)
        Please provide all periods.
        >>> 'periods_order' in view.getSessionData()
        False

    A much more likely scenario is that some period was selected
    twice, and some was missed out:

        >>> request = TestRequest(form={'period_0': 'D', 'period_1': 'C',
        ...                             'period_2': 'C'})
        >>> view = PeriodOrderSimple(context, request)
        >>> view.getSessionData()['period_names'] = ['A', 'B', 'C', 'D']
        >>> view.getSessionData()['time_slots'] = time_slots
        >>> view.getSessionData()['day_names'] = ['1', '2']
        >>> view.update()
        False
        >>> print translate(view.error)
        The following periods were selected more than once: C
        >>> 'periods_order' in view.getSessionData()
        False

    The view will display the order the user has selected though:

        >>> print view()
        <BLANKLINE>
        ...
          <div class="row">
            <select name="period_0">
              <option>A</option>
              <option>B</option>
              <option>C</option>
              <option selected="selected">D</option>
            </select>
          </div>
          <div class="row">
            <select name="period_1">
              <option>A</option>
              <option>B</option>
              <option selected="selected">C</option>
              <option>D</option>
            </select>
          </div>
          <div class="row">
            <select name="period_2">
              <option>A</option>
              <option>B</option>
              <option selected="selected">C</option>
              <option>D</option>
            </select>
          </div>
          ...

    """


def doctest_wizard_PeriodOrderComplex():
    """Unit test for PeriodOrderComplex view

        >>> from schooltool.timetable.browser.ttwizard import \\
        ...     PeriodOrderComplex
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = PeriodOrderComplex(context, request)

    Let's say we have some periods and days in a rotating cycle:

        >>> session = view.getSessionData()
        >>> session['period_names'] = ['A', 'B', 'C', 'D']
        >>> session['day_names'] = ['Day One', 'Day Two']
        >>> session['cycle'] = 'rotating'
        >>> session['time_model'] = 'rotating'

    The number of period dropdowns is the maximum of slots in a day:

        >>> time_slots = [
        ...     [(datetime.time(9, 30), datetime.timedelta(0, 3300))],
        ...     [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
        ...      (datetime.time(10, 35), datetime.timedelta(0, 2700)),
        ...      (datetime.time(10, 35), datetime.timedelta(0, 2700)),]]
        >>> session['time_slots'] = time_slots

    Our view lets the template easily access period and days names:

        >>> view.periods()
        ['A', 'B', 'C', 'D']

        >>> view.days()
        ['Day One', 'Day Two']

    The number of dropdowns for a day is equal to the number of slots:

        >>> view.numSlots()
        [1, 3]

    If we render the view, we get a list of dropdowns with consecutive
    periods selected:

        >>> print view()
        <BLANKLINE>
        ...
          <table>
            <tr>
              <th>Day One</th>
              <th>Day Two</th>
            </tr>
            <tr>
              <td>
                <select name="period_0_0">
                  <option selected="selected">A</option>
                  <option>B</option>
                  <option>C</option>
                  <option>D</option>
                </select>
              </td>
              <td>
                <select name="period_1_0">
                  <option selected="selected">A</option>
                  <option>B</option>
                  <option>C</option>
                  <option>D</option>
                </select>
              </td>
            </tr>
            <tr>
              <td>
              </td>
              <td>
                <select name="period_1_1">
                  <option>A</option>
                  <option selected="selected">B</option>
                  <option>C</option>
                  <option>D</option>
                </select>
              </td>
            </tr>
            <tr>
              <td>
              </td>
              <td>
                <select name="period_1_2">
                  <option>A</option>
                  <option>B</option>
                  <option selected="selected">C</option>
                  <option>D</option>
                </select>
              </td>
            </tr>
          </table>
        ...

    When the user shuffles the dropdowns and submits them, the order
    of periods is changed:

        >>> request = TestRequest(form={'period_0_0': 'A',
        ...                             'period_1_0': 'B', 'period_1_1': 'A',
        ...                             'period_1_2': 'C'})
        >>> view = PeriodOrderComplex(context, request)
        >>> view.getSessionData()['period_names'] = ['A', 'B', 'C', 'D']
        >>> view.getSessionData()['day_names'] = ['Day 1', 'Day 2']
        >>> view.getSessionData()['time_slots'] = time_slots
        >>> view.getSessionData()['cycle'] = 'rotating'
        >>> view.getSessionData()['time_model'] = 'rotating'
        >>> view.update()
        True
        >>> print view.getSessionData()['periods_order']
        [['A'], ['B', 'A', 'C']]

    The next step is always the homeroom step:

        >>> view.next()
        <class 'schooltool.timetable.browser.ttwizard.HomeroomStep'>

    If not all periods are in the request, update fails and the user
    gets an error.  This is unlikely in real life as the dropdowns
    have values preselected:

        >>> request = TestRequest(form={'period_0_0': 'A', 'period_0_1': 'B',
        ...                             'period_1_1': 'A'})
        >>> view = PeriodOrderComplex(context, request)
        >>> view.getSessionData()['period_names'] = ['A', 'B', 'C', 'D']
        >>> view.getSessionData()['day_names'] = ['Z', 'X']
        >>> view.getSessionData()['time_slots'] = time_slots
        >>> view.getSessionData()['cycle'] = 'rotating'
        >>> view.getSessionData()['time_model'] = 'rotating'
        >>> view.update()
        False
        >>> print translate(view.error)
        Please provide all periods.
        >>> 'periods_order' not in view.getSessionData()
        True

    A much more likely scenario is that some period was selected
    twice, and some was missed out:

        >>> request = TestRequest(form={'period_0_0': 'B', 'period_0_1': 'B',
        ...                             'period_1_2': 'C',
        ...                             'period_1_0': 'A', 'period_1_1': 'A'})
        >>> view = PeriodOrderComplex(context, request)
        >>> view.getSessionData()['period_names'] = ['A', 'B', 'C', 'D']
        >>> view.getSessionData()['day_names'] = ['X', 'Y']
        >>> view.getSessionData()['time_slots'] = [
        ...     [(datetime.time(9, 30), datetime.timedelta(0, 3300)),
        ...      (datetime.time(10, 30), datetime.timedelta(0, 3300))],
        ...     [(datetime.time(9, 15), datetime.timedelta(0, 3300)),
        ...      (datetime.time(10, 35), datetime.timedelta(0, 2700)),
        ...      (datetime.time(10, 35), datetime.timedelta(0, 2700)),]]
        >>> view.getSessionData()['cycle'] = 'rotating'
        >>> view.getSessionData()['time_model'] = 'rotating'
        >>> view.update()
        False
        >>> print translate(view.error)
        The following periods were selected more than once: B on day X, A on day Y
        >>> 'periods_order' not in view.getSessionData()
        True

    The view will display the order the user has selected though:

        >>> print view()
        <BLANKLINE>
        ...
          <div class="error">The following periods were selected more
            than once: B on day X, A on day Y</div>
          <table>
            <tr>
              <th>X</th>
              <th>Y</th>
            </tr>
            <tr>
              <td>
                <select name="period_0_0">
                  <option>A</option>
                  <option selected="selected">B</option>
                  <option>C</option>
                  <option>D</option>
                </select>
              </td>
              <td>
                <select name="period_1_0">
                  <option selected="selected">A</option>
                  <option>B</option>
                  <option>C</option>
                  <option>D</option>
                </select>
              </td>
            </tr>
            <tr>
              <td>
                <select name="period_0_1">
                  <option>A</option>
                  <option selected="selected">B</option>
                  <option>C</option>
                  <option>D</option>
                </select>
              </td>
              <td>
                <select name="period_1_1">
                  <option selected="selected">A</option>
                  <option>B</option>
                  <option>C</option>
                  <option>D</option>
                </select>
              </td>
            </tr>
            <tr>
              <td>
              </td>
              <td>
                <select name="period_1_2">
                  <option>A</option>
                  <option>B</option>
                  <option selected="selected">C</option>
                  <option>D</option>
                </select>
              </td>
            </tr>
          </table>
        ...

    """


def doctest_wizard_PeriodOrderComplex_weekly_rotating():
    """Unit test for PeriodOrderComplex view

        >>> from schooltool.timetable.browser.ttwizard import \\
        ...     PeriodOrderComplex
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = PeriodOrderComplex(context, request)

    There is one special case: a rotating cycle that uses weekdays to determine
    time slots.

        >>> session = view.getSessionData()
        >>> session['cycle'] = 'rotating'
        >>> session['time_model'] = 'weekly'

    Let's say we have some periods:

        >>> session['period_names'] = ['A', 'B', 'C', 'D']
        >>> session['day_names'] = ['Day One', 'Day Two']

    Our view lets the template easily access them:

        >>> view.periods()
        ['A', 'B', 'C', 'D']

        >>> view.days()
        ['Day One', 'Day Two']

    The number of dropdowns for each day is equal to the number of periods:

        >>> view.numSlots()
        [4, 4]

    """


def doctest_wizard_HomeroomStep():
    r"""Unit test for HomeroomStep

        >>> from schooltool.timetable.browser.ttwizard import HomeroomStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = HomeroomStep(context, request)

    If you say that your school does not take attendance in a homeroom period,
    you will be redirected to FinalStep.

        >>> session = view.getSessionData()
        >>> session['homeroom'] = False
        >>> view.next()
        <...ttwizard.FinalStep...>

    Otherwise, you will have to specify the homeroom period for each day

        >>> session['homeroom'] = True
        >>> view.next()
        <...ttwizard.HomeroomPeriodsStep...>

    """


def doctest_wizard_HomeroomPeriodsStep():
    r"""Unit test for HomeroomPeriodsStep view

        >>> from schooltool.timetable.browser.ttwizard \
        ...         import HomeroomPeriodsStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = HomeroomPeriodsStep(context, request)

    Let's say we have some days with periods:

        >>> view.getSessionData()['day_names'] = ['Day1', 'Day2', 'Day3']
        >>> view.getSessionData()['periods_order'] = [['A', 'B', 'C', 'D'],
        ...                                           ['B', 'C', 'D', 'E'],
        ...                                           ['C', 'D', 'E', 'F']]

    Our view lets the template easily access them:

        >>> view.days()
        ['Day1', 'Day2', 'Day3']

        >>> view.periodsInOrder()
        [['A', 'B', 'C', 'D'],
         ['B', 'C', 'D', 'E'],
         ['C', 'D', 'E', 'F']]

    If we render the view, we get lists of checkboxes with period names,
    one list for each day.

        >>> print view()
        <BLANKLINE>
        ...
              <td>
        <BLANKLINE>
        <BLANKLINE>
                    <input type="checkbox" name="homeroom_0_A"
                           id="homeroom_0_A" />
                    <label for="homeroom_0_A">A</label>
        <BLANKLINE>
                  <br />
        <BLANKLINE>
        <BLANKLINE>
        <BLANKLINE>
                    <input type="checkbox" name="homeroom_0_B"
                           id="homeroom_0_B" />
                    <label for="homeroom_0_B">B</label>
        <BLANKLINE>
                  <br />
        ...

    When the user selects some of the items and submits the form, the choice
    is remembered.

        >>> request = TestRequest(form={'homeroom_0_A': 'checked',
        ...                             'homeroom_0_C': 'checked',
        ...                             'homeroom_2_C': 'checked',
        ...                             'homeroom_2_F': 'checked'})
        >>> view = HomeroomPeriodsStep(context, request)
        >>> view.getSessionData()['day_names'] = ['Day 1', 'Day 2', 'Day 3']
        >>> view.getSessionData()['periods_order'] = [['A', 'B', 'C', 'D'],
        ...                                           ['B', 'C', 'D', 'E'],
        ...                                           ['C', 'D', 'E', 'F']]

        >>> view.update()
        True
        >>> print view.getSessionData()['homeroom_periods']
        [['A', 'C'], [], ['C', 'F']]

    The next step is always the final step:

        >>> view.next()
        <class 'schooltool.timetable.browser.ttwizard.FinalStep'>

    """


def doctest_wizard_FinalStep():
    r"""Unit test for FinalStep

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = FinalStep(context, request)

        >>> from datetime import time, timedelta
        >>> data = view.getSessionData()
        >>> data['title'] = u'Sample Timetable'
        >>> data['cycle'] = 'weekly'
        >>> data['day_names'] = ['Monday', 'Tuesday', 'Wednesday',
        ...                      'Thursday', 'Friday']
        >>> data['similar_days'] = True
        >>> data['time_slots'] = [[(time(9, 30), timedelta(minutes=55)),
        ...                        (time(10, 30), timedelta(minutes=55))]]
        >>> data['named_periods'] = False
        >>> from schooltool.timetable.browser.ttwizard \
        ...     import default_period_names
        >>> data['periods_order'] = default_period_names(data['time_slots'])
        >>> data['homeroom'] = False

        >>> view()

        >>> timetable = context['sample-timetable']
        >>> timetable
        <...Timetable object at ...>

        >>> print timetable.title
        Sample Timetable

    We should get redirected to the timetables index:

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/schoolyear/timetables'

    The cycle of steps loops here

        >>> view.update()
        True
        >>> view.next()
        <...FirstStep...>

    """

def doctest_wizard_FinalStep_createTimeSlots():
    """Unit test for FinalStep.createTimeSlots

        >>> from datetime import time, timedelta
        >>> from schooltool.timetable.timetable import Timetable
        >>> from schooltool.timetable.browser.ttwizard import FinalStep

        >>> class FinalStepForTest(FinalStep):
        ...     def addTimeSlotTemplates(self, schedule, days):
        ...         for key, title, time_slots in days:
        ...             slots = [format_time_range(*slot) for slot in time_slots]
        ...             print '%s, %s: %s' % (key, title, slots)

        >>> day_names = ['Day one', 'Day two', 'Day three']
        >>> slots = [[(time(9, 0), timedelta(minutes=50)),
        ...           (time(12, 35), timedelta(minutes=50))],
        ...          [(time(10, 0), timedelta(minutes=50))],
        ...          [(time(11, 0), timedelta(minutes=50)),
        ...           (time(14, 35), timedelta(minutes=50))],
        ...         ]

        >>> context = timetables
        >>> request = TestRequest()

    Weekly model replaces day names with weekday names.

        >>> view = FinalStepForTest(context, request)
        >>> timetable = Timetable(None, None, title=u'Default')

        >>> print timetable.time_slots
        None

        >>> view.createTimeSlots(timetable, 'weekly', day_names, slots)
        0, Monday: ['09:00-09:50', '12:35-13:25']
        1, Tuesday: ['10:00-10:50']
        2, Wednesday: ['11:00-11:50', '14:35-15:25']

        >>> print timetable.time_slots
        <schooltool.timetable.daytemplates.WeekDayTemplates object ...>

    It can create up to 7 days timeslots.

        >>> view = FinalStepForTest(context, request)
        >>> timetable = Timetable(None, None, title=u'Default')

        >>> many_slots = slots * 4
        >>> len(many_slots)
        12

        >>> view.createTimeSlots(timetable, 'weekly', day_names, many_slots)
        0, Monday: ['09:00-09:50', '12:35-13:25']
        1, Tuesday: ['10:00-10:50']
        2, Wednesday: ['11:00-11:50', '14:35-15:25']
        3, Thursday: ['09:00-09:50', '12:35-13:25']
        4, Friday: ['10:00-10:50']
        5, Saturday: ['11:00-11:50', '14:35-15:25']
        6, Sunday: ['09:00-09:50', '12:35-13:25']

    Rotating model uses the given day titles.

        >>> view = FinalStepForTest(context, request)
        >>> timetable = Timetable(None, None, title='Default')

        >>> view.createTimeSlots(timetable, 'rotating', day_names, slots)
        0, Day one: ['09:00-09:50', '12:35-13:25']
        1, Day two: ['10:00-10:50']
        2, Day three: ['11:00-11:50', '14:35-15:25']

        >>> print timetable.time_slots
        <schooltool.timetable.daytemplates.SchoolDayTemplates object ...>

    We do not support unknown models.

        >>> view.createTimeSlots(timetable, 'this is bogus', day_names, slots)
        Traceback (most recent call last):
        ...
        NotImplementedError

    """


def doctest_wizard_FinalStep_createPeriods():
    """Unit test for FinalStep.createPeriods

        >>> from datetime import time, timedelta
        >>> from schooltool.timetable.timetable import Timetable
        >>> from schooltool.timetable.browser.ttwizard import FinalStep

        >>> class FinalStepForTest(FinalStep):
        ...     def addPeriodTemplates(self, schedule, days):
        ...         for key, title, periods in days:
        ...             print '%s, %s: %s' % (
        ...                 key, title,
        ...                 ['%s %s' % (act_type, period)
        ...                  for period, act_type in periods])

        >>> day_names = ['Day one', 'Day two', 'Day three']
        >>> period_names = [
        ...     ['first', 'second'],
        ...     ['first'],
        ...     ['first', 'second', 'third'],
        ...     ]
        >>> homeroom_names = [
        ...     ['second'],
        ...     [],
        ...     ['first', 'third'],
        ...     ]

        >>> context = timetables
        >>> request = TestRequest()

    Weekly model creates a weekly period template schedule.

        >>> view = FinalStepForTest(context, request)
        >>> timetable = Timetable(None, None, title=u'Default')

        >>> print timetable.periods
        None

        >>> view.createPeriods(timetable, 'weekly', day_names,
        ...                    period_names, homeroom_names)
        0, Day one: ['lesson first', 'homeroom second']
        1, Day two: ['lesson first']
        2, Day three: ['homeroom first', 'lesson second', 'homeroom third']

        >>> print timetable.periods
        <schooltool.timetable.daytemplates.WeekDayTemplates object ...>

    Weekly model supports up to 7 days.

        >>> view = FinalStepForTest(context, request)
        >>> timetable = Timetable(None, None, title=u'Default')

        >>> many_days = ['Day %s' % chr(n+ord('A')) for n in range(10)]
        >>> many_periods = [['X']] * len(many_days)

        >>> view.createPeriods(timetable, 'weekly', many_days,
        ...                    many_periods, [()]*len(many_periods))
        0, Day A: ['lesson X']
        1, Day B: ['lesson X']
        2, Day C: ['lesson X']
        3, Day D: ['lesson X']
        4, Day E: ['lesson X']
        5, Day F: ['lesson X']
        6, Day G: ['lesson X']

    Rotating model is also quite simple.

        >>> view = FinalStepForTest(context, request)
        >>> timetable = Timetable(None, None, title=u'Default')

        >>> print timetable.periods
        None

        >>> view.createPeriods(timetable, 'rotating', day_names,
        ...                    period_names, homeroom_names)
        0, Day one: ['lesson first', 'homeroom second']
        1, Day two: ['lesson first']
        2, Day three: ['homeroom first', 'lesson second', 'homeroom third']

        >>> print timetable.periods
        <schooltool.timetable.daytemplates.SchoolDayTemplates object ...>

    We do not support unknown models.

        >>> view.createPeriods(timetable, 'bogus', day_names,
        ...                    period_names, homeroom_names)
        Traceback (most recent call last):
        ...
        NotImplementedError

    """


def doctest_wizard_FinalStep_createTimetable():
    """Unit tests for FinalStep.createTimetable

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = FinalStep(context, request)


    This method creates a new timetable object.

        >>> data = view.getSessionData()
        >>> data['title'] = u'Some timetable'

        >>> timetable = view.createTimetable()

        >>> print timetable.title
        Some timetable

    The timetable's timezone is 'UTC', the app default:

        >>> timetable.timezone
        'UTC'

    However if we set a school preferred timezone and try again:

        >>> IApplicationPreferences(app).timezone = 'Australia/Canberra'

        >>> timetable = view.createTimetable()

        >>> timetable.timezone
        'Australia/Canberra'

    """


def doctest_wizard_FinalStep_setUpTimetable():
    r"""Unit test for FinalStep.setUpTimetable

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = FinalStep(context, request)
        >>> data = view.getSessionData()
        >>> data['title'] = u'Default'

    First, let's try the simple weekly model.

        >>> data['cycle'] = 'weekly'
        >>> data['day_names'] = ['Monday', 'Tuesday', 'Wednesday',
        ...                      'Thursday', 'Friday']
        >>> from datetime import time, timedelta
        >>> data['similar_days'] = True
        >>> data['time_slots'] = [[(time(9, 30), timedelta(minutes=55)),
        ...                        (time(10, 30), timedelta(minutes=55))]] * 5
        >>> data['named_periods'] = False
        >>> from schooltool.timetable.browser.ttwizard \
        ...     import default_period_names
        >>> data['periods_order'] = default_period_names(data['time_slots'])
        >>> data['homeroom'] = False

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (WeekDayTemplates)
        +--------------------+--------------------+--------------------+--------------------+--------------------+
        | Monday             | Tuesday            | Wednesday          | Thursday           | Friday             |
        +--------------------+--------------------+--------------------+--------------------+--------------------+
        | lesson 09:30-10:25 | lesson 09:30-10:25 | lesson 09:30-10:25 | lesson 09:30-10:25 | lesson 09:30-10:25 |
        | lesson 10:30-11:25 | lesson 10:30-11:25 | lesson 10:30-11:25 | lesson 10:30-11:25 | lesson 10:30-11:25 |
        +--------------------+--------------------+--------------------+--------------------+--------------------+
        Time slots (WeekDayTemplates)
        +-------------+-------------+-------------+-------------+-------------+
        | Monday      | Tuesday     | Wednesday   | Thursday    | Friday      |
        +-------------+-------------+-------------+-------------+-------------+
        | 09:30-10:25 | 09:30-10:25 | 09:30-10:25 | 09:30-10:25 | 09:30-10:25 |
        | 10:30-11:25 | 10:30-11:25 | 10:30-11:25 | 10:30-11:25 | 10:30-11:25 |
        +-------------+-------------+-------------+-------------+-------------+

    The model can also be rotating.

        >>> data['cycle'] = 'rotating'
        >>> data['day_names'] = ['D1', 'D2', 'D3']

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (SchoolDayTemplates)
        +--------------------+--------------------+--------------------+
        | D1                 | D2                 | D3                 |
        +--------------------+--------------------+--------------------+
        | lesson 09:30-10:25 | lesson 09:30-10:25 | lesson 09:30-10:25 |
        | lesson 10:30-11:25 | lesson 10:30-11:25 | lesson 10:30-11:25 |
        +--------------------+--------------------+--------------------+
        Time slots (SchoolDayTemplates)
        +-------------+-------------+-------------+
        | D1          | D2          | D3          |
        +-------------+-------------+-------------+
        | 09:30-10:25 | 09:30-10:25 | 09:30-10:25 |
        | 10:30-11:25 | 10:30-11:25 | 10:30-11:25 |
        +-------------+-------------+-------------+

        >>> print_day_templates(timetable.periods.templates)
        +--------------------+--------------------+--------------------+
        | D1                 | D2                 | D3                 |
        +--------------------+--------------------+--------------------+
        | lesson 09:30-10:25 | lesson 09:30-10:25 | lesson 09:30-10:25 |
        | lesson 10:30-11:25 | lesson 10:30-11:25 | lesson 10:30-11:25 |
        +--------------------+--------------------+--------------------+

        >>> print_day_templates(timetable.time_slots.templates)
        +-------------+-------------+-------------+
        | D1          | D2          | D3          |
        +-------------+-------------+-------------+
        | 09:30-10:25 | 09:30-10:25 | 09:30-10:25 |
        | 10:30-11:25 | 10:30-11:25 | 10:30-11:25 |
        +-------------+-------------+-------------+

    The periods can be named rather than be designated by time:

        >>> data['cycle'] = 'rotating'
        >>> data['day_names'] = ['D1', 'D2', 'D3']
        >>> data['named_periods'] = True
        >>> data['period_names'] = ['Green', 'Blue']
        >>> data['periods_same'] = True
        >>> data['periods_order'] = [['Green', 'Blue']] * 3

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (SchoolDayTemplates)
        +--------------+--------------+--------------+
        | D1           | D2           | D3           |
        +--------------+--------------+--------------+
        | lesson Green | lesson Green | lesson Green |
        | lesson Blue  | lesson Blue  | lesson Blue  |
        +--------------+--------------+--------------+
        Time slots (SchoolDayTemplates)
        +-------------+-------------+-------------+
        | D1          | D2          | D3          |
        +-------------+-------------+-------------+
        | 09:30-10:25 | 09:30-10:25 | 09:30-10:25 |
        | 10:30-11:25 | 10:30-11:25 | 10:30-11:25 |
        +-------------+-------------+-------------+

    """


def doctest_wizard_FinalStep_setUpTimetable_different_order_on_different_days_weekly():
    """Unit test for FinalStep.setUpTimetable

    Weekly cycle, same time slots on each day, different period order in
    each day.

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> view = FinalStep(timetables, TestRequest())
        >>> data = view.getSessionData()

        >>> from datetime import time, timedelta
        >>> slots = [(time(8+n, 0), timedelta(minutes=45))
        ...          for n in range(4)]

        >>> data['title'] = u'Default'
        >>> data['cycle'] = 'weekly'
        >>> data['day_names'] = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
        ...                      'Friday']
        >>> data['similar_days'] = True
        >>> data['time_slots'] = [slots] * 5
        >>> data['named_periods'] = True
        >>> data['period_names'] = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
        >>> data['periods_same'] = False
        >>> data['periods_order'] = [['A', 'B', 'C', 'D'],
        ...                          ['B', 'C', 'D', 'E'],
        ...                          ['C', 'D', 'E', 'F'],
        ...                          ['D', 'E', 'F', 'G'],
        ...                          ['E', 'F', 'G', 'H']]
        >>> data['homeroom'] = False

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (WeekDayTemplates)
        +----------+----------+-----------+----------+----------+
        | Monday   | Tuesday  | Wednesday | Thursday | Friday   |
        +----------+----------+-----------+----------+----------+
        | lesson A | lesson B | lesson C  | lesson D | lesson E |
        | lesson B | lesson C | lesson D  | lesson E | lesson F |
        | lesson C | lesson D | lesson E  | lesson F | lesson G |
        | lesson D | lesson E | lesson F  | lesson G | lesson H |
        +----------+----------+-----------+----------+----------+
        Time slots (WeekDayTemplates)
        +-------------+-------------+-------------+-------------+-------------+
        | Monday      | Tuesday     | Wednesday   | Thursday    | Friday      |
        +-------------+-------------+-------------+-------------+-------------+
        | 08:00-08:45 | 08:00-08:45 | 08:00-08:45 | 08:00-08:45 | 08:00-08:45 |
        | 09:00-09:45 | 09:00-09:45 | 09:00-09:45 | 09:00-09:45 | 09:00-09:45 |
        | 10:00-10:45 | 10:00-10:45 | 10:00-10:45 | 10:00-10:45 | 10:00-10:45 |
        | 11:00-11:45 | 11:00-11:45 | 11:00-11:45 | 11:00-11:45 | 11:00-11:45 |
        +-------------+-------------+-------------+-------------+-------------+

    """


def doctest_wizard_FinalStep_setUpTimetable_different_order_on_different_days_cyclic():
    """Unit test for FinalStep.setUpTimetable

    Rotating cycle, same time slots on each day, different period order in
    each day.

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> view = FinalStep(timetables, TestRequest())
        >>> data = view.getSessionData()

        >>> from datetime import time, timedelta
        >>> slots = [(time(8+n, 0), timedelta(minutes=45))
        ...          for n in range(4)]

        >>> data['title'] = u'Default'
        >>> data['cycle'] = 'rotating'
        >>> data['day_names'] = ['Day 1', 'Day 2', 'Day 3']
        >>> data['similar_days'] = True
        >>> data['time_slots'] = [slots] * 3
        >>> data['named_periods'] = True
        >>> data['period_names'] = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> data['periods_same'] = False
        >>> data['periods_order'] = [['A', 'B', 'C', 'D'],
        ...                          ['B', 'C', 'D', 'E'],
        ...                          ['C', 'D', 'E', 'F']]
        >>> data['homeroom'] = False

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (SchoolDayTemplates)
        +----------+----------+----------+
        | Day 1    | Day 2    | Day 3    |
        +----------+----------+----------+
        | lesson A | lesson B | lesson C |
        | lesson B | lesson C | lesson D |
        | lesson C | lesson D | lesson E |
        | lesson D | lesson E | lesson F |
        +----------+----------+----------+
        Time slots (SchoolDayTemplates)
        +-------------+-------------+-------------+
        | Day 1       | Day 2       | Day 3       |
        +-------------+-------------+-------------+
        | 08:00-08:45 | 08:00-08:45 | 08:00-08:45 |
        | 09:00-09:45 | 09:00-09:45 | 09:00-09:45 |
        | 10:00-10:45 | 10:00-10:45 | 10:00-10:45 |
        | 11:00-11:45 | 11:00-11:45 | 11:00-11:45 |
        +-------------+-------------+-------------+

    """


def doctest_wizard_FinalStep_setUpTimetable_different_order_cyclic_weekly():
    """Unit test for FinalStep.setUpTimetable

    Rotating cycle, different time slots on a weekly basis, different period
    order in each day.

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> view = FinalStep(timetables, TestRequest())
        >>> data = view.getSessionData()

        >>> from datetime import time, timedelta
        >>> slots = [[(time(8+n, d*5), timedelta(minutes=45))
        ...           for n in range(4)]
        ...          for d in range(5)]

        >>> data['title'] = u'Default'
        >>> data['cycle'] = 'rotating'
        >>> data['day_names'] = ['Day 1', 'Day 2', 'Day 3']
        >>> data['similar_days'] = False
        >>> data['time_model'] = 'weekly'
        >>> data['time_slots'] = slots
        >>> data['named_periods'] = True
        >>> data['period_names'] = ['A', 'B', 'C', 'D', 'E', 'F']
        >>> data['periods_same'] = False
        >>> data['periods_order'] = [['A', 'B', 'C', 'D', 'E', 'F'],
        ...                          ['B', 'C', 'D', 'E', 'F', 'A'],
        ...                          ['C', 'D', 'E', 'F', 'A', 'B']]
        >>> data['homeroom'] = False

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (SchoolDayTemplates)
        +----------+----------+----------+
        | Day 1    | Day 2    | Day 3    |
        +----------+----------+----------+
        | lesson A | lesson B | lesson C |
        | lesson B | lesson C | lesson D |
        | lesson C | lesson D | lesson E |
        | lesson D | lesson E | lesson F |
        | lesson E | lesson F | lesson A |
        | lesson F | lesson A | lesson B |
        +----------+----------+----------+
        Time slots (WeekDayTemplates)
        +-------------+-------------+-------------+-------------+-------------+
        | Monday      | Tuesday     | Wednesday   | Thursday    | Friday      |
        +-------------+-------------+-------------+-------------+-------------+
        | 08:00-08:45 | 08:05-08:50 | 08:10-08:55 | 08:15-09:00 | 08:20-09:05 |
        | 09:00-09:45 | 09:05-09:50 | 09:10-09:55 | 09:15-10:00 | 09:20-10:05 |
        | 10:00-10:45 | 10:05-10:50 | 10:10-10:55 | 10:15-11:00 | 10:20-11:05 |
        | 11:00-11:45 | 11:05-11:50 | 11:10-11:55 | 11:15-12:00 | 11:20-12:05 |
        +-------------+-------------+-------------+-------------+-------------+

    """


def doctest_wizard_FinalStep_setUpTimetable_different_times():
    """Unit test for FinalStep.setUpTimetable

    Weekly cycle, different time slots on each day, different period order in
    each day.

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> view = FinalStep(timetables, TestRequest())
        >>> data = view.getSessionData()

        >>> from datetime import time, timedelta
        >>> slots = [[(time(8+n, d*5), timedelta(minutes=45))
        ...           for n in range(4)]
        ...          for d in range(5)]

        >>> data['title'] = u'Default'
        >>> data['cycle'] = 'weekly'
        >>> data['day_names'] = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
        ...                      'Friday']
        >>> data['similar_days'] = False
        >>> data['time_slots'] = slots
        >>> data['named_periods'] = True
        >>> data['period_names'] = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
        >>> data['periods_same'] = False
        >>> data['periods_order'] = [['A', 'B', 'C', 'D'],
        ...                          ['B', 'C', 'D', 'E'],
        ...                          ['C', 'D', 'E', 'F'],
        ...                          ['D', 'E', 'F', 'G'],
        ...                          ['E', 'F', 'G', 'H']]
        >>> data['homeroom'] = False

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (WeekDayTemplates)
        +----------+----------+-----------+----------+----------+
        | Monday   | Tuesday  | Wednesday | Thursday | Friday   |
        +----------+----------+-----------+----------+----------+
        | lesson A | lesson B | lesson C  | lesson D | lesson E |
        | lesson B | lesson C | lesson D  | lesson E | lesson F |
        | lesson C | lesson D | lesson E  | lesson F | lesson G |
        | lesson D | lesson E | lesson F  | lesson G | lesson H |
        +----------+----------+-----------+----------+----------+
        Time slots (WeekDayTemplates)
        +-------------+-------------+-------------+-------------+-------------+
        | Monday      | Tuesday     | Wednesday   | Thursday    | Friday      |
        +-------------+-------------+-------------+-------------+-------------+
        | 08:00-08:45 | 08:05-08:50 | 08:10-08:55 | 08:15-09:00 | 08:20-09:05 |
        | 09:00-09:45 | 09:05-09:50 | 09:10-09:55 | 09:15-10:00 | 09:20-10:05 |
        | 10:00-10:45 | 10:05-10:50 | 10:10-10:55 | 10:15-11:00 | 10:20-11:05 |
        | 11:00-11:45 | 11:05-11:50 | 11:10-11:55 | 11:15-12:00 | 11:20-12:05 |
        +-------------+-------------+-------------+-------------+-------------+

    """


def doctest_wizard_FinalStep_setUpTimetable_with_homeroom():
    """Unit test for FinalStep.setUpTimetable

    Weekly cycle, different time slots on each day, different period order in
    each day, homeroom periods.

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> view = FinalStep(timetables, TestRequest())
        >>> data = view.getSessionData()

        >>> from datetime import time, timedelta
        >>> slots = [[(time(8+n, d*5), timedelta(minutes=45))
        ...           for n in range(4)]
        ...          for d in range(5)]

        >>> data['title'] = u'Default'
        >>> data['cycle'] = 'weekly'
        >>> data['day_names'] = ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
        ...                      'Friday']
        >>> data['similar_days'] = False
        >>> data['time_slots'] = slots
        >>> data['named_periods'] = True
        >>> data['period_names'] = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']
        >>> data['periods_same'] = False
        >>> data['periods_order'] = [['A', 'B', 'C', 'D'],
        ...                          ['B', 'C', 'D', 'E'],
        ...                          ['C', 'D', 'E', 'F'],
        ...                          ['D', 'E', 'F', 'G'],
        ...                          ['E', 'F', 'G', 'H']]
        >>> data['homeroom'] = True
        >>> data['homeroom_periods'] = [['A'], ['C'], [], ['F'], ['H']]

        >>> timetable = Timetable(None, None, title=u'Default')
        >>> view.setUpTimetable(timetable)

        >>> print_timetable(timetable)
        Timetable 'Default'
        Periods (WeekDayTemplates)
        +------------+------------+-----------+------------+------------+
        | Monday     | Tuesday    | Wednesday | Thursday   | Friday     |
        +------------+------------+-----------+------------+------------+
        | homeroom A | lesson B   | lesson C  | lesson D   | lesson E   |
        | lesson B   | homeroom C | lesson D  | lesson E   | lesson F   |
        | lesson C   | lesson D   | lesson E  | homeroom F | lesson G   |
        | lesson D   | lesson E   | lesson F  | lesson G   | homeroom H |
        +------------+------------+-----------+------------+------------+
        Time slots (WeekDayTemplates)
        +-------------+-------------+-------------+-------------+-------------+
        | Monday      | Tuesday     | Wednesday   | Thursday    | Friday      |
        +-------------+-------------+-------------+-------------+-------------+
        | 08:00-08:45 | 08:05-08:50 | 08:10-08:55 | 08:15-09:00 | 08:20-09:05 |
        | 09:00-09:45 | 09:05-09:50 | 09:10-09:55 | 09:15-10:00 | 09:20-10:05 |
        | 10:00-10:45 | 10:05-10:50 | 10:10-10:55 | 10:15-11:00 | 10:20-11:05 |
        | 11:00-11:45 | 11:05-11:50 | 11:10-11:55 | 11:15-12:00 | 11:20-12:05 |
        +-------------+-------------+-------------+-------------+-------------+

    """


def doctest_wizard_FinalStep_add():
    """Unit test for FinalStep.setUpTimetable

        >>> from schooltool.timetable.browser.ttwizard import FinalStep
        >>> context = timetables
        >>> request = TestRequest()
        >>> view = FinalStep(context, request)

        >>> from schooltool.timetable.timetable import Timetable
        >>> timetable = Timetable(None, None, title='timetable')
        >>> view.add(timetable)

        >>> context['timetable'] is timetable
        True

    """


def doctest_wizard_TimetableWizard():
    """Unit test for TimetableWizard

        >>> context = timetables
        >>> request = TestRequest()
        >>> view = TimetableWizard(context, request)

    We shall stub it heavily.

        >>> class StepStub:
        ...     update_succeeds = False
        ...     def __init__(self, context, request):
        ...         pass
        ...     def __repr__(self):
        ...         return '<same step>'
        ...     def update(self):
        ...         print 'Updating...'
        ...         return self.update_succeeds
        ...     def __call__(self):
        ...         return 'Rendered step'
        ...     def next(self):
        ...         return NextStepStub

        >>> class NextStepStub(object):
        ...     def __init__(self, context, request):
        ...         pass
        ...     def __repr__(self):
        ...         return '<next step>'
        ...     def __call__(self):
        ...         return 'Rendered next step'

        >>> view.getLastStep = lambda: StepStub(None, None)

        >>> def rememberLastStep(step):
        ...     print 'Remembering step: %s' % step
        >>> view.rememberLastStep = rememberLastStep

    There are three main cases.

    Case 1: the user completes the current step successfully.

        >>> StepStub.update_succeeds = True
        >>> print view()
        Updating...
        Remembering step: <next step>
        Rendered next step

    Case 2: the user does not complete the current step successfully.

        >>> StepStub.update_succeeds = False
        >>> print view()
        Updating...
        Remembering step: <same step>
        Rendered step

    Case 3: the user presses the 'Cancel' button

        >>> view.request.form['CANCEL'] = 'Cancel'
        >>> view()
        Remembering step: <...FirstStep...>

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/schoolyear/timetables'

    """


def doctest_wizard_TimetableWizard_getLastStep():
    """Unit test for TimetableWizard.getLastStep

        >>> context = timetables
        >>> request = TestRequest()
        >>> view = TimetableWizard(context, request)

    When there is no step saved in the session, getLastStep returns the first
    step.

        >>> view.getLastStep()
        <...FirstStep...>

    When there is one, that's what getLastStep returns.

        >>> from schooltool.timetable.browser.ttwizard import CycleStep
        >>> view.getSessionData()['last_step'] = CycleStep
        >>> view.getLastStep()
        <...CycleStep...>

    """


def doctest_wizard_TimetableWizard_rememberLastStep():
    """Unit test for TimetableWizard.rememberLastStep

        >>> context = timetables
        >>> request = TestRequest()
        >>> view = TimetableWizard(context, request)

        >>> from schooltool.timetable.browser.ttwizard import CycleStep
        >>> view.rememberLastStep(CycleStep(context, request))
        >>> view.getSessionData()['last_step']
        <class 'schooltool.timetable.browser.ttwizard.CycleStep'>

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return unittest.TestSuite([
                doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                     optionflags=optionflags),
                doctest.DocTestSuite('schooltool.timetable.browser.ttwizard',
                                     optionflags=optionflags),
           ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for report functionality.

"""
import unittest
import doctest

from zope.app.testing import setup

from schooltool.report import report


def doctest_Something(self):
    """Tests for something.
    """


def setUp(test=None):
    setup.placefulSetUp()


def tearDown(test=None):
    setup.placefulTearDown()


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown)
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 32.

Evolution script to fix a typo in default 'ethnicity' demographics field.
"""
from zope.app.generations.utility import findObjectsProviding
from zope.app.publication.zopepublication import ZopePublication

from schooltool.basicperson.interfaces import IEnumFieldDescription
from schooltool.app.interfaces import ISchoolToolApplication


DEMOGRAPHICS_FIELDS_KEY = 'schooltool.basicperson.demographics_fields'


def evolve(context):
    root = context.connection.root().get(ZopePublication.root_name, None)

    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        if DEMOGRAPHICS_FIELDS_KEY not in app:
            continue
        fields = app[DEMOGRAPHICS_FIELDS_KEY]
        for field in fields.values():
            if not IEnumFieldDescription.providedBy(field):
                continue
            for n, item in enumerate(field.items):
                if item == u'Native Hawaiian or Other Pasific Islander':
                    field.items[n] = u'Native Hawaiian or Other Pacific Islander'




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve32
"""
import unittest
import doctest

from pprint import pprint
from zope.app.testing import setup
from zope.interface import implements
from zope.container import btree

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.generations.tests import ContextStub
import schooltool.basicperson.interfaces
import schooltool.basicperson.demographics
from schooltool import basicperson


class AppStub(btree.BTreeContainer):
    implements(ISchoolToolApplication)


def doctest_evolve32():
    """Evolution to generation 32.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()

    Create demographics with the typo (assume that user deleted or otherwise
    modified some of the demographics).

        >>> fields = basicperson.demographics.DemographicsFields()
        >>> app['schooltool.basicperson.demographics_fields'] = fields

        >>> fields['ID'] = basicperson.demographics.TextFieldDescription(
        ...     'ID', 'ID')
        >>> fields['birth'] = basicperson.demographics.TextFieldDescription(
        ...     'birth', 'Place of birth')

        >>> fields['ethnic'] = basicperson.demographics.EnumFieldDescription(
        ...     'ethnic', 'Ethnicity')
        >>> fields['ethnic'].items = [
        ...     u'Outter space',
        ...     u'Native Hawaiian or Other Pasific Islander',
        ...     u'Asian']

    Let's evolve now.

        >>> from schooltool.generations.evolve32 import evolve
        >>> evolve(context)

    The typo is fixed.

        >>> pprint(fields['ethnic'].items)
        [u'Outter space',
         u'Native Hawaiian or Other Pacific Islander',
         u'Asian']

    """


def setUp(test):
    setup.placelessSetUp()
    setup.setUpAnnotations()

def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Resource views
"""
from collections import defaultdict
from datetime import datetime

import z3c.form
import z3c.form.browser.text
import zc.table.table
from z3c.form import form, field, button, widget
from z3c.form.interfaces import DISPLAY_MODE, HIDDEN_MODE, NO_VALUE
from zc.table.column import GetterColumn

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapter, adapts
from zope.component import getUtilitiesFor
from zope.component import queryAdapter
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.component import getMultiAdapter
from zope.container.interfaces import INameChooser
from zope.event import notify
from zope.formlib import form as oldform
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.interface import implementer, implements, implementsOnly
from zope.lifecycleevent import ObjectCreatedEvent
from zope.publisher.browser import BrowserView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.security.checker import canAccess
from zope.session.interfaces import ISession
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.browser.app import EditRelationships
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.basicperson.browser.demographics import (
    DemographicsView,
    FlourishDemographicsView, FlourishReorderDemographicsView)
from schooltool.basicperson.interfaces import IAddEditViewTitle
from schooltool.basicperson.interfaces import ILimitKeysLabel
from schooltool.basicperson.interfaces import ILimitKeysHint
from schooltool.common.inlinept import InheritTemplate, InlineViewPageTemplate
from schooltool.resource.interfaces import IBookingCalendar
from schooltool.resource.interfaces import (
             IResourceContainer, IResourceTypeInformation, IResourceSubTypes,
             IResource, IEquipment, ILocation, IResourceDemographicsFields)
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.resource.resource import Resource, Location, Equipment
from schooltool.person.browser.person import PersonFilterWidget
from schooltool.resource.interfaces import IResourceFactoryUtility
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.skin import flourish
from schooltool import table

from schooltool.common import SchoolToolMessage as _


class ResourceContainerView(oldform.FormBase):
    """A Resource Container view."""

    __used_for__ = IResourceContainer

    index_title = _("Resource index")

    prefix = "resources"
    form_fields = oldform.Fields()
    searchActions = oldform.Actions(
        oldform.Action('Search', success='handle_search_action'),)

    actions = oldform.Actions(
                oldform.Action('Delete', success='handle_delete_action'),
                oldform.Action('Book', success='handle_book_action'))
    template = ViewPageTemplateFile("resourcecontainer.pt")
    delete_template = ViewPageTemplateFile('container_delete.pt')

    resourceType = None

    def __init__(self, context, request):
        oldform.FormBase.__init__(self, context, request)
        self.resourceType = self.request.get('SEARCH_TYPE','|').split('|')[0]
        self.filter_widget = queryMultiAdapter(
            (self.getResourceUtility(), self.request),
            table.interfaces.IFilterWidget)


    def getSubTypes(self):
        utilities = sorted(getUtilitiesFor(IResourceFactoryUtility))

        types = []
        for name, utility in utilities:
            if IResourceSubTypes.providedBy(utility):
                subTypeAdapter = utility
            else:
                subTypeAdapter = queryAdapter(utility, IResourceSubTypes,
                                          default=None)
                if subTypeAdapter != None:
                    subTypeAdapter = subTypeAdapter()

            typeHeader = [name, utility.title, 'unclickable']
            subTypes = subTypeAdapter.types()
            if subTypes:
                types.append(typeHeader)
                for subtype in subTypes:
                    types.append([name, subtype, 'clickable'])
        return types

    def listIdsForDeletion(self):
        return [key for key in self.context
                if "delete.%s" % key in self.request]

    def _listItemsForDeletion(self):
        return [self.context[key] for key in self.listIdsForDeletion()]

    itemsToDelete = property(_listItemsForDeletion)

    def handle_delete_action(self, action, data):
        self.template = self.delete_template

    def handle_search_action(self, action, data):
        pass

    def handle_book_action(self, action, data):
        sessionWrapper = ISession(self.request)
        session = sessionWrapper['schooltool.resource']
        session['bookingSelection'] = [''.join(key.split('.')[1:])
                                       for key in self.request
                                       if key.startswith('delete.')]
        url = absoluteURL(IBookingCalendar(self.context), self.request)
        return self.request.response.redirect(url)

    def getResourceUtility(self):
        return queryUtility(IResourceFactoryUtility,
                            name=self.resourceType, default=None)

    def columns(self):
        return self.getResourceUtility().columns()

    def sortOn(self):
        return (('title', False), )

    def filter(self, values):
        return self.filter_widget.filter(values)

    def renderResourceTable(self):
        columns = [table.column.CheckboxColumn(
                       prefix="delete", name='delete', title=u'')]
        available_columns = self.columns()
        available_columns[0].cell_formatter = table.table.url_cell_formatter

        columns.extend(available_columns)
        formatter = zc.table.table.StandaloneFullFormatter(
            self.context, self.request, self.filter(self.context.values()),
            columns=columns,
            sort_on=self.sortOn(),
            prefix="available")
        formatter.cssClasses['table'] = 'data'
        return formatter()


class FlourishResourcesView(flourish.page.Page):

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/table" />
    ''')


class ResourcesTable(table.ajax.Table):

    def columns(self):
        default = table.ajax.Table.columns(self)
        description = GetterColumn(
            name='description',
            title=_(u'Description'),
            getter=lambda i, f: i.description or '')
        return default + [description]


class ResourceListTable(ResourcesTable):

    @property
    def source(self):
        return ISchoolToolApplication(None)['resources']

    def items(self):
        return self.context


class LocationListTable(ResourceListTable):

    def items(self):
        return [r for r in self.context
                if ILocation(r, None) is not None]


class EquipmentListTable(ResourceListTable):

    def items(self):
        return [r for r in self.context
                if IEquipment(r, None) is not None]


class EditResourceRelationships(EditRelationships):

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['resources']


class EditLocationRelationships(EditResourceRelationships):
    pass


class EditEquipmentRelationships(EditResourceRelationships):
    pass


class LocationAddRelationshipTable(RelationshipAddTableMixin,
                                   ResourcesTable):
    pass


class LocationRemoveRelationshipTable(RelationshipRemoveTableMixin,
                                      ResourcesTable):
    pass


class EquipmentAddRelationshipTable(RelationshipAddTableMixin,
                                    ResourcesTable):
    pass


class EquipmentRemoveRelationshipTable(RelationshipRemoveTableMixin,
                                       ResourcesTable):
    pass


class BaseTypeFilter(table.table.FilterWidget):
    """Base Type Filter"""

    def render(self):
        if 'CLEAR_SEARCH' in self.request:
            self.request.form['SEARCH'] = ''
        return self.template()

    def filter(self, list):
        resourceType = self.request.get('SEARCH_TYPE','|').split('|')[0]
        list = [resource for resource in list
                 if IResourceTypeInformation(resource).id == resourceType]
        if 'SEARCH' in self.request and 'CLEAR_SEARCH' not in self.request:
            searchstr = self.request['SEARCH'].lower()
            results = [item for item in list
                       if (searchstr in item.title.lower() and
                           self.request.get('SEARCH_TYPE','|').split('|')[1] == item.type)]
        elif 'SEARCH_TYPE' in self.request:
            results = [item for item in list
                       if self.request.get('SEARCH_TYPE','|').split('|')[1] == item.type]
        else:
            results = list

        return results


class EquipmentTypeFilter(BaseTypeFilter):
    """Equipment Type Filter"""


class LocationTypeFilter(BaseTypeFilter):
    """Location Type Filter"""


class ResourceTypeFilter(BaseTypeFilter):
    """Resource Type Filter"""


class IResourceSubTypeWidget(z3c.form.interfaces.ITextWidget):
    pass


class ResourceSubTypeWidget(z3c.form.browser.text.TextWidget):
    implementsOnly(IResourceSubTypeWidget)

    utility = None

    def __init__(self, request, utility=None):
        super(ResourceSubTypeWidget, self).__init__(request)
        self.utility = utility

    def freeTextValue(self):
        if self.value in self.subTypes():
            return ''
        return self.value

    def subTypes(self):
        util = queryUtility(IResourceFactoryUtility,
                            name=self.utility,
                            default=None)
        if IResourceSubTypes.providedBy(util):
            subtypes = util
        else:
            subtypes = queryAdapter(util, IResourceSubTypes, default=None)
        if subtypes is not None:
            return subtypes.types()
        return []

    def hasInput(self):
        return self.request.get(self.name,None) != '' or self.request.get(self.name+'.newSubType',None)

    def extract(self, default=NO_VALUE):
        subType = self.request.get(self.name, default)
        newSubType = self.request.get(self.name+'.newSubType', default)
        if subType and subType != default:
            return subType
        return newSubType or default


def ResourceSubTypeFieldWidget(field, request):
    utility_name = 'resource'
    if not field.interface is None:
        # XXX: this is what we get for using named utilities
        if issubclass(field.interface, IEquipment):
            utility_name = 'equipment'
        elif issubclass(field.interface, ILocation):
            utility_name = 'location'
    return widget.FieldWidget(
        field,
        ResourceSubTypeWidget(request, utility=utility_name)
        )


class ResourceContainerFilterWidget(PersonFilterWidget):

    template = ViewPageTemplateFile('resource_filter.pt')
    parameters = ['SEARCH_TITLE', 'SEARCH_TYPE']

    def types(self):
        utilities = sorted(getUtilitiesFor(IResourceFactoryUtility))
        types = []
        for name, utility in utilities:
            if IResourceSubTypes.providedBy(utility):
                subTypeAdapter = utility
            else:
                subTypeAdapter = queryAdapter(utility, IResourceSubTypes,
                                          default=None)
                if subTypeAdapter != None:
                    subTypeAdapter = subTypeAdapter()

            typeHeader = {'id':name,
                          'title':utility.title,
                          'clickable':False}
            types.append(typeHeader)
            for subtype in subTypeAdapter.types():
                types.append({'id':name, 'title':subtype, 'clickable':True})
        return types


    def filter(self, list):
        if 'CLEAR_SEARCH' in self.request:
            for parameter in self.parameters:
                self.request.form[parameter] = ''
            return list

        results = list

        if 'SEARCH_TITLE' in self.request:
            searchstr = self.request['SEARCH_TITLE'].lower()
            results = [item for item in results
                       if searchstr in item.title.lower()]

        if 'SEARCH_TYPE' in self.request:
            type = self.request['SEARCH_TYPE']
            if not type:
                return results
            type = type.split('|')[0]
            utility = queryUtility(IResourceFactoryUtility,
                                   name=type, default=None)
            filter_widget = queryMultiAdapter((utility,self.request),
                                              table.interfaces.IFilterWidget)
            if filter_widget:
                results = filter_widget.filter(results)

        return results


class FlourishResourceContainerFilterWidget(ResourceContainerFilterWidget):

    template = ViewPageTemplateFile('templates/f_resource_filter.pt')

    search_title_id = 'SEARCH_TITLE'
    search_type_id = 'SEARCH_TYPE'

    def types(self):
        options = [
            {'id': 'equipment',
             'title': _('Equipment')},
            {'id': 'location',
             'title': _('Location')},
            {'id': 'resource',
             'title': _('Resource')},
            ]
        return options

    def filter(self, results):
        if self.search_title_id in self.request:
            searchstr = self.request[self.search_title_id].lower()
            results = [item for item in results
                       if searchstr in item.title.lower() or
                       (item.description and searchstr in item.description.lower())]
        if self.search_type_id in self.request:
            type = self.request[self.search_type_id]
            if not type:
                return results
            results = [resource for resource in results
                       if IResourceTypeInformation(resource).id == type]
        return results


class ResourcesTableFilter(table.ajax.TableFilter,
                           FlourishResourceContainerFilterWidget):

    template = ViewPageTemplateFile('templates/f_resource_table_filter.pt')

    @property
    def search_title_id(self):
        return self.manager.html_id+"-title"

    @property
    def search_group_id(self):
        return self.manager.html_id+"-group"

    def filter(self, results):
        if self.ignoreRequest:
            return results
        return FlourishResourceContainerFilterWidget.filter(self, results)


class ResourceContainerLinks(flourish.page.RefineLinksViewlet):
    """Resource container links viewlet."""


class ResourceImportLinks(flourish.page.RefineLinksViewlet):
    """Resource import links viewlet."""


class ResourceDemographicsFieldsAbsoluteURLAdapter(BrowserView):

    adapts(IResourceDemographicsFields, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = str(getMultiAdapter((app, self.request), name='absolute_url'))
        return url + '/resource_demographics'

    __call__ = __str__


class ResourceDemographicsView(DemographicsView):

    title = _('Resource Attributes')


class FlourishResourceDemographicsView(FlourishDemographicsView):

    container_class = 'container'

    keys = [('resource', _('Res.')),
            ('location', _('Locat.')),
            ('equipment', _('Equip.'))]


class FlourishReorderResourceDemographicsView(FlourishReorderDemographicsView):
    pass


@adapter(IResourceDemographicsFields)
@implementer(IAddEditViewTitle)
def getAddEditViewTitle(context):
    return _('Resource attributes')


@adapter(IResourceDemographicsFields)
@implementer(ILimitKeysLabel)
def getLimitKeysLabel(context):
    return _('Limit to resource type(s)')


@adapter(IResourceDemographicsFields)
@implementer(ILimitKeysHint)
def getLimitKeysHint(context):
    return _(u"If you select one or more resource types below, this field "
              "will only be displayed in forms and reports for "
              "resources of the selected types.")


##########  Base class of all resource views (uses self.resource_type) #########
class ResourceFieldGenerator(object):

    def makeRows(self, fields, cols=1):
        rows = []
        while fields:
            rows.append(fields[:cols])
            fields = fields[cols:]
        return rows

    def makeFieldSet(self, fieldset_id, legend, fields, cols=1):
        result = {
            'id': fieldset_id,
            'legend': legend,
            }
        result['rows'] = self.makeRows(fields, cols)
        return result

    def fieldsets(self):
        result = []
        sources = [
            (self.base_id, self.base_legend, list(self.getBaseFields())),
            (self.demo_id, self.demo_legend, list(self.getDemoFields())),
            ]
        for fieldset_id, legend, fields in sources:
            result.append(self.makeFieldSet(fieldset_id, legend, fields, 2))
        return result

    def getDemoFields(self):
        fields = field.Fields()
        dfs = IResourceDemographicsFields(ISchoolToolApplication(None))
        for field_desc in dfs.filter_key(self.resource_type):
            fields += field_desc.makeField()
        return fields


###############  Resource view (group determined by base class) #################
class BaseResourceView(form.Form, ResourceFieldGenerator):

    template = ViewPageTemplateFile('templates/resource_view.pt')
    mode = DISPLAY_MODE
    id = 'resource-view'

    @property
    def label(self):
        return self.context.title

    def update(self):
        self.fields = self.getBaseFields()
        self.fields += self.getDemoFields()
        self.subforms = []
        super(BaseResourceView, self).update()

    def __call__(self):
        self.update()
        return self.render()

    def updateWidgets(self):
        super(BaseResourceView, self).updateWidgets()
        for widget in self.widgets:
            if not self.widgets[widget].value:
                self.widgets[widget].mode = HIDDEN_MODE


class FlourishBaseResourceView(flourish.page.Page, BaseResourceView):

    def update(self):
        BaseResourceView.update(self)

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')

    def has_leaders(self):
        return bool(list(self.context.leaders))


class ResourceView(BaseResourceView):
    """A location info view."""

    resource_type = 'resource'

    def getBaseFields(self):
        return field.Fields(IResource)


class FlourishResourceView(FlourishBaseResourceView, ResourceView):

    def getBaseFields(self):
        return field.Fields(IResource).select('title', 'description')


class LocationView(BaseResourceView):
    """A location info view."""

    resource_type = 'location'

    def getBaseFields(self):
        return field.Fields(ILocation)


class FlourishLocationView(FlourishBaseResourceView, LocationView):

    def getBaseFields(self):
        return field.Fields(ILocation).select('title', 'description')


class EquipmentView(BaseResourceView):
    """A equipment info view."""

    resource_type = 'equipment'

    def getBaseFields(self):
        return field.Fields(IEquipment)


class FlourishEquipmentView(FlourishBaseResourceView, EquipmentView):

    def getBaseFields(self):
        return field.Fields(IEquipment).select('title', 'description')


###############  Base classes of all resource add/edit views ################
# XXX: move this to a generic form base class
#      it's also duplicated in basicperson.browser.person.PersonForm
class ErrorMessageBase(object):

    formErrorsMessage = _('Please correct the marked fields below.')


class BaseResourceAddView(ErrorMessageBase, form.AddForm,
                          ResourceFieldGenerator):

    id = 'resource-form'
    template = ViewPageTemplateFile('templates/resource_form.pt')

    def getBaseFields(self):
        return field.Fields(IResource)

    def updateActions(self):
        super(BaseResourceAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Add'))
    def handleAdd(self, action):
        super(BaseResourceAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


    def createAndAdd(self, data):
        resource = self._factory()
        resource.title = data.get('title')
        chooser = INameChooser(self.context)
        resource.__name__ = chooser.chooseName('', resource)
        form.applyChanges(self, resource, data)
        notify(ObjectCreatedEvent(resource))
        self.context[resource.__name__] = resource
        return resource

    def update(self):
        self.fields = self.getBaseFields()
        self.fields += self.getDemoFields()
        self.updateWidgets()
        self.updateActions()
        self.actions.execute()

    def updateWidgets(self):
        super(BaseResourceAddView, self).updateWidgets()

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class BaseResourceEditView(ErrorMessageBase, form.EditForm,
                           ResourceFieldGenerator):

    id = 'resource-form'
    template = ViewPageTemplateFile('templates/resource_form.pt')

    def update(self):
        self.fields = self.getBaseFields()
        self.fields += self.getDemoFields()
        super(BaseResourceEditView, self).update()

    @button.buttonAndHandler(_('Apply'))
    def handle_apply_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        self.applyChanges(data)
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    @property
    def label(self):
        return _(u'Change information for ${fullname}',
                 mapping={'fullname': self.context.title})

    def updateActions(self):
        super(BaseResourceEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class BaseFlourishResourceAddForm(flourish.form.AddForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None

    def createAndAdd(self, data):
        resource = self._factory()
        resource.title = data.get('title')
        chooser = INameChooser(self.context)
        resource.__name__ = chooser.chooseName('', resource)
        form.applyChanges(self, resource, data)
        notify(ObjectCreatedEvent(resource))
        self.context[resource.__name__] = resource
        self._resource = resource
        return resource

    def nextURL(self):
        return absoluteURL(getattr(self, '_resource', None) or self.context,
                           self.request)


###############  Resource add/edit views ################
class BaseResourceForm(object):

    resource_type = 'resource'
    base_id = 'base-data'
    base_legend = _('Resource identification')
    demo_id = 'demo-data'
    demo_legend = _('Resource attributes')

    def getBaseFields(self):
        return field.Fields(IResource).omit('type', 'notes')


class ResourceAddView(BaseResourceForm, BaseResourceAddView):

    label = _('Add new resource')
    _factory = Resource


class FlourishResourceAddView(BaseFlourishResourceAddForm, ResourceAddView):
    pass


class ResourceEditView(BaseResourceForm, BaseResourceEditView):

    label = _('Edit resource')


class FlourishResourceEditView(flourish.page.Page, ResourceEditView):

    label = None

    def update(self):
        ResourceEditView.update(self)


###############  Location add/edit views ################
class BaseLocationForm(object):

    resource_type = 'location'
    base_id = 'base-data'
    base_legend = _('Location identification')
    demo_id = 'demo-data'
    demo_legend = _('Location attributes')

    def getBaseFields(self):
        fields = field.Fields(ILocation).select('title', 'description')
        return fields


class LocationAddView(BaseLocationForm, BaseResourceAddView):

    label = _('Add new location')
    _factory = Location


class FlourishLocationAddView(BaseFlourishResourceAddForm, LocationAddView):
    pass


class LocationEditView(BaseLocationForm, BaseResourceEditView):

    label = _('Edit location')


class FlourishLocationEditView(flourish.page.Page, LocationEditView):

    label = None

    def update(self):
        LocationEditView.update(self)


###############  Equipment add/edit views ################
class BaseEquipmentForm(object):

    resource_type = 'equipment'
    base_id = 'base-data'
    base_legend = _('Equipment identification')
    demo_id = 'demo-data'
    demo_legend = _('Equipment attributes')

    def getBaseFields(self):
        fields = field.Fields(IEquipment).select('title', 'description')
        return fields


class EquipmentAddView(BaseEquipmentForm, BaseResourceAddView):

    label = _('Add new equipment')
    _factory = Equipment


class FlourishEquipmentAddView(BaseFlourishResourceAddForm, EquipmentAddView):
    pass


class EquipmentEditView(BaseEquipmentForm, BaseResourceEditView):

    label = _('Edit equipment')


class FlourishEquipmentEditView(flourish.page.Page, EquipmentEditView):

    label = None

    def update(self):
        EquipmentEditView.update(self)


class FlourishResourceContainerTableFormatter(table.table.SchoolToolTableFormatter):

    def makeFormatter(self):
        formatter = table.table.SchoolToolTableFormatter.makeFormatter(self)
        formatter.cssClasses['table'] = 'resources-table'
        return formatter


class ResourceLinks(flourish.page.RefineLinksViewlet):

    @property
    def title(self):
        return self.context.title


class ResourceActions(flourish.page.RefineLinksViewlet): pass


class FlourishBookResourceView(BrowserView):

    def __call__(self):
        sessionWrapper = ISession(self.request)
        session = sessionWrapper['schooltool.resource']
        session['bookingSelection'] = [self.context.__name__]
        url = absoluteURL(IBookingCalendar(self.context), self.request)
        self.request.response.redirect(url)


class FlourishResourceDeleteView(flourish.form.DialogForm, form.Form):

    dialog_submit_actions = ('delete',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_('Delete'))
    def handle_delete_action(self, action):
        parent = self.context.__parent__
        parent_url = absoluteURL(parent, self.request)
        context_name = self.context.__name__.encode('utf-8')
        url = '%s/delete.html?delete.%s&CONFIRM' % (parent_url, context_name)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishResourceDeleteView, self).updateActions()
        self.actions['delete'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishResourceDeleteLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        title = _(u'Delete ${resource_title}',
                  mapping={'resource_title': self.context.title})
        return translate(title, context=self.request)


class FlourishManageResourcesOverview(flourish.page.Content,
                                      ActiveSchoolYearContentMixin):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_resources_overview.pt')

    @property
    def resources(self):
        return self.context['resources']

    @Lazy
    def resource_types(self):
        types = defaultdict(lambda:dict(amount=0, title=None, id=None))
        for resource in self.resources.values():
            info = IResourceTypeInformation(resource)
            if types[info.id]['title'] is None:
                types[info.id]['title'] = info.title
            if types[info.id]['id'] is None:
                types[info.id]['id'] = info.id
            types[info.id]['amount'] += 1
        return types

    def resources_url(self):
        return self.url_with_schoolyear_id(self.context,
                                           view_name='resources')


class FlourishRequestResourceReportView(RequestRemoteReportDialog):

    report_builder='resource_report.pdf'


class ResourceReportPDFView(flourish.report.PlainPDFPage):

    name = _("Resource Report")

    message_title = _("resource report")

    @property
    def title(self):
        preferences = IApplicationPreferences(ISchoolToolApplication(None))
        return preferences.title

    def makeFileName(self, basename):
        timestamp = datetime.now().strftime('%y%m%d%H%M')
        return '%s_%s.pdf' % (basename, timestamp)

    @property
    def filename(self):
        return self.makeFileName('resource_report')

    def ptos(self):
        collator = ICollator(self.request.locale)
        ptos_dict = {}
        for resource in sorted(self.context.values(),
                               key=lambda x:collator.key(x.title)):
            if ILocation(resource, None) is not None:
                resource_type = _('Location')
            elif IEquipment(resource, None) is not None:
                resource_type = _('Equipment')
            else:
                resource_type = _('Resource')
            pto = ptos_dict.setdefault(resource_type, {
                'type': resource_type,
                'rows': [],
                })
            pto['rows'].append({
                'title': resource.title,
                'description': resource.description,
                })
        return [ptos_dict[k] for k in sorted(ptos_dict.keys(),
                                             key=lambda x:collator.key(x))]




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Many-to-many relationships.

See the README.txt file in this package for a detailed overview with working
code examples.
"""

from schooltool.relationship.uri import URIObject, IURIObject       # reexport
from schooltool.relationship.relationship import relate             # reexport
from schooltool.relationship.relationship import unrelate           # reexport
from schooltool.relationship.relationship import unrelateAll        # reexport
from schooltool.relationship.relationship import getRelatedObjects  # reexport
from schooltool.relationship.relationship import RelationshipSchema # reexport
from schooltool.relationship.relationship import RelationshipProperty # ditto



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
ZCML directives for reports

"""

from zope.configuration.fields import MessageID
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserView
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.schema import TextLine
from zope.viewlet.metadirectives import IViewletDirective

from schooltool.skin import flourish
from schooltool.report.interfaces import IReportLinkViewletManager
from schooltool.report.report import ReportLinkViewlet
from schooltool.report.report import getReportRegistrationUtility
from schooltool.skin.flourish.zcml import IViewletOrder


class IReportLinkDirective(IViewletDirective, IViewletOrder):
    group = MessageID(title=u"Report group", required=True)
    title = MessageID(title=u"Report link text", required=True)
    description = MessageID(title=u"Report link description", required=True)
    file_type = TextLine(title=u"Report file extension", required=False)
    link = TextLine(title=u"Link to the report", required=False)


def reportLinkDirective(_context, name, permission, for_=Interface,
    layer=IDefaultBrowserLayer, view=IBrowserView,
    manager=IReportLinkViewletManager,
    class_=ReportLinkViewlet, template=None, group=u'', title=u'',
    description='', link='', file_type='', **kwargs):

    flourish.zcml.viewlet(
        _context, name, permission,
        for_=for_, layer=layer, view=view,
        manager=manager,
        class_=class_, template=template,
        group=group, title=title, description=description, link=link,
        file_type=file_type, **kwargs)

    # Register the report for reference
    utility = getReportRegistrationUtility()
    utility.registerReport(group, title, description, file_type, name, layer)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Period and time slot schedule templates.
Template scheduling over dates.
"""

from persistent import Persistent
from zope.interface import implements, implementer
from zope.component import adapter
from zope.container.contained import Contained
from zope.container.ordered import OrderedContainer
from zope.container.contained import containedEvent
from zope.event import notify

from schooltool.common import DateRange
from schooltool.timetable import interfaces


class DayTemplate(OrderedContainer):
    implements(interfaces.IDayTemplate)

    title = None

    def __init__(self, title=u''):
        OrderedContainer.__init__(self)
        self.title = title


class DayTemplateContainer(OrderedContainer):
    implements(interfaces.IDayTemplateContainer)


class TimeSlot(Persistent, Contained):
    implements(interfaces.ITimeSlot)

    tstart = None
    duration = None
    activity_type = None

    def __init__(self, tstart, duration, activity_type=None):
        Contained.__init__(self)
        self.tstart = tstart
        self.duration = duration
        self.activity_type = activity_type

    def __cmp__(self, other):
        return cmp((self.tstart, self.duration),
                   (other.tstart, other.duration))


class DayTemplateSchedule(Persistent, Contained):
    """Day templates scheduled by date."""
    implements(interfaces.IDayTemplateSchedule)

    templates = None

    def initTemplates(self):
        self.templates, event = containedEvent(
            DayTemplateContainer(), self, 'templates')
        notify(event)

    def iterDates(self, dates):
        for date in dates:
            yield None


class CalendarDayTemplates(DayTemplateSchedule):
    implements(interfaces.ICalendarDayTemplates)

    starting_index = 0

    def getDay(self, schedule, date):
        assert self.templates
        days_passed = (date - schedule.first).days
        keys = self.templates.keys()
        n = (self.templates.starting_index + days_passed) % len(keys)
        return self.templates[keys[n]]

    def iterDates(self, dates):
        if not self.templates:
            for date in dates:
                yield None
            return
        schedule = interfaces.ISchedule(self)
        scheduled_dates = DateRange(schedule.first, schedule.last)
        for date in dates:
            if date not in scheduled_dates:
                yield None
            else:
                day = self.getDay(schedule, date)
                yield day


class WeekDayTemplates(DayTemplateSchedule):
    implements(interfaces.IWeekDayTemplates)

    def getWeekDayKey(self, weekday):
        return unicode(weekday)

    def getWeekDay(self, weekday):
        return self.templates.get(unicode(weekday), None)

    def iterDates(self, dates):
        if not self.templates:
            for date in dates:
                yield None
            return
        schedule = interfaces.ISchedule(self)
        schooldays = interfaces.ISchooldays(schedule)
        scheduled_dates = DateRange(schedule.first, schedule.last)
        for date in dates:
            if (date not in scheduled_dates or
                date not in schooldays):
                yield None
            else:
                day = self.getWeekDay(date.weekday())
                yield day


class SchoolDayTemplates(DayTemplateSchedule):
    implements(interfaces.ISchoolDayTemplates)

    starting_index = 0

    def getDayIndex(self, schedule, schooldays, date):
        assert self.templates
        day_index = self.starting_index
        if date == schedule.first:
            return day_index

        if date > schedule.first:
            skip_dates = DateRange(schedule.first, date - date.resolution)
        else:
            skip_dates = DateRange(date + date.resolution, schedule.first)
        skipped_schooldays = len(list(schooldays.iterDates(skip_dates)))
        if date < schedule.first:
            skipped_schooldays = -skipped_schooldays

        day_index = (day_index + skipped_schooldays) % len(self.templates)
        return day_index

    def iterDates(self, dates):
        if not self.templates:
            for date in dates:
                yield None
            return
        schedule = interfaces.ISchedule(self)
        scheduled_dates = DateRange(schedule.first, schedule.last)
        schooldays = interfaces.ISchooldays(schedule)
        prev_index = None
        prev_date = None
        for date in dates:
            if (prev_date is not None and
                date - prev_date != date.resolution):
                # This is not the next day, reset day index
                prev_index = None

            if (date not in scheduled_dates or
                date not in schooldays):
                # Not a scheduled schoolday
                yield None
            else:
                if prev_index is None:
                    day_index = self.getDayIndex(schedule, schooldays, date)
                else:
                    day_index = (prev_index + 1) % len(self.templates)
                keys = self.templates.keys()
                yield self.templates[keys[day_index]]
                prev_index = day_index
            prev_date = date


@adapter(interfaces.IDayTemplateSchedule)
@implementer(interfaces.ISchedule)
def getScheduledTemplatesSchedule(day_templates):
    return day_templates.__parent__



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for basic person.
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.component import provideAdapter
from zope.interface import implements

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences


def doctest_BasicPerson():
    """Tests for BasicPerson.

         >>> from zope.interface.verify import verifyObject
         >>> from schooltool.basicperson.person import BasicPerson
         >>> person = BasicPerson("peter", "Peter", "Johnson")

         >>> from schooltool.basicperson.interfaces import IBasicPerson
         >>> verifyObject(IBasicPerson, person)
         True

         >>> person.title
         'Peter Johnson'

    """


def doctest_PersonFactoryUtility():
    """Tests for PersonFactoryUtility.

        >>> from schooltool.basicperson.person import PersonFactoryUtility
        >>> factory = PersonFactoryUtility()

        >>> from schooltool.person.interfaces import IPersonFactory
        >>> from zope.interface.verify import verifyObject
        >>> verifyObject(IPersonFactory, factory)
        True

        >>> class AppStub(object):
        ...     implements(ISchoolToolApplication, IApplicationPreferences)
        ...     def __init__(self):
        ...         self.name_sorting = 'last_name'
        >>> app = AppStub()
        >>> provideAdapter(lambda _: app,
        ...                adapts=(None, ), provides=ISchoolToolApplication)

        >>> for column in factory.columns():
        ...     print "%s, %s" % (column.name, column.title)
        last_name, Last Name
        first_name, First Name

        >>> factory.sortOn()
        (('last_name', False), ('first_name', False))

    If we change name sorting order, both the sort and column order changes:

        >>> app.name_sorting = 'first_name'
        >>> for column in factory.columns():
        ...     print "%s, %s" % (column.name, column.title)
        first_name, First Name
        last_name, Last Name

        >>> factory.sortOn()
        (('first_name', False), ('last_name', False))

    """


def doctest_PersonFactoryUtility_createManagerUser():
    """Tests for PersonFactoryUtility.createManagerUser

    First let's create the utility:

        >>> from schooltool.basicperson.person import PersonFactoryUtility
        >>> utility = PersonFactoryUtility()

    The title of the manager user is set to "Administratorius" + system name:

        >>> manager = utility.createManagerUser("manager_username")
        >>> manager.title
        'Default Manager'
        >>> manager.username
        'manager_username'

    """


def setUp(test):
    setup.placelessSetUp()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    return doctest.DocTestSuite(optionflags=optionflags,
                                setUp=setUp, tearDown=tearDown)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Person Preferences implementation
"""

from persistent import Persistent
from zope.interface import implements
from zope.annotation.interfaces import IAnnotations
from zope.location.location import ILocation

from schooltool.securitypolicy.crowds import OwnerCrowd
from schooltool.securitypolicy.crowds import ConfigurableCrowd
from schooltool.person import interfaces

PERSON_PREFERENCES_KEY = 'schooltool.app.PersonPreferences'


class PersonPreferences(Persistent):

    implements(interfaces.IPersonPreferences, ILocation)

    __name__ = 'preferences'
    __parent__ = None

    # XXX: Only available in schooltool, but that's okay for now.
    cal_periods = True
    cal_public = False


def getPersonPreferences(person):
    """Adapt an IAnnotatable object to IPersonPreferences."""
    annotations = IAnnotations(person)
    try:
        return annotations[PERSON_PREFERENCES_KEY]
    except KeyError:
        prefs = PersonPreferences()
        prefs.__parent__ = person
        annotations[PERSON_PREFERENCES_KEY] = prefs
        return prefs


def getPreferencesOwner(preferences):
    """Adapt IPersonPreferences to IPerson"""
    return interfaces.IPerson(preferences.__parent__)


class PersonPreferencesEditorsCrowd(ConfigurableCrowd):

    setting_key = 'persons_can_set_their_preferences'

    def contains(self, principal):
        """Return the value of the related setting (True or False)."""
        return (ConfigurableCrowd.contains(self, principal) and
                OwnerCrowd(self.context).contains(principal))



# -*- coding: utf-8 -*-
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for group views.
"""
import unittest
import doctest

from zope.publisher.browser import TestRequest

from schooltool.app.browser.testing import setUp, tearDown


def doctest_ResourceView():
    r"""Test for ResourceView

    Let's create a view for a resource:

        >>> from schooltool.resource.browser.resource import ResourceView
        >>> from schooltool.resource.resource import Resource
        >>> resource = Resource()
        >>> request = TestRequest()
        >>> view = ResourceView(resource, request)

    """


def doctest_ResourceCSVImporter():
    r"""Tests for ResourceCSVImporter.

    Create a resource container and an importer

        >>> from schooltool.resource.browser.csvimport import \
        ...     ResourceCSVImporter
        >>> from schooltool.resource.resource import ResourceContainer
        >>> container = ResourceContainer()
        >>> importer = ResourceCSVImporter(container, None)

    Import some sample data

        >>> csvdata='''Resource 1, Resource 1 Description
        ... Resource2
        ... Resource3, Resource 3 Description, location
        ... Resource4, , location'''
        >>> importer.importFromCSV(csvdata)
        True

    Check that the resources exist

        >>> [resource for resource in container]
        [u'resource-1', u'resource2', u'resource3', u'resource4']

    Check that descriptions were imported properly

        >>> from schooltool.resource.interfaces import ILocation
        >>> [resource.description for resource in container.values()]
        ['Resource 1 Description', '', 'Resource 3 Description', '']
        >>> [ILocation.providedBy(resource) for resource in container.values()]
        [False, False, True, True]

    """


def doctest_ResourceCSVImportView():
    r"""
    We'll create a resource csv import view

        >>> from schooltool.resource.browser.csvimport import \
        ...     ResourceCSVImportView
        >>> from schooltool.resource.resource import ResourceContainer
        >>> container = ResourceContainer()
        >>> request = TestRequest()

    Now we'll try a text import.  Note that the description is not required

        >>> request.form = {'csvtext' : u"A Resource, The best Resource\nAnother Resource\nRecurso en Español, Spanish resource\n\n\n",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = ResourceCSVImportView(container, request)
        >>> view.update()
        >>> [resource for resource in container]
        [u'a-resource', u'another-resource', u'recurso-en-espa\xe3ol']

    If no data is provided, we naturally get an error

        >>> request.form = {'charset' : 'UTF-8', 'UPDATE_SUBMIT': 1}
        >>> view = ResourceCSVImportView(container, request)
        >>> view.update()
        >>> view.errors
        [u'No data provided']

    We also get an error if a line starts with a comma (no title)

        >>> request.form = {'csvtext' : u", No title provided here",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = ResourceCSVImportView(container, request)
        >>> view.update()
        >>> view.errors
        [u'Failed to import CSV text', u'Titles may not be empty']

    """


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=doctest.ELLIPSIS|
                                                   doctest.REPORT_NDIFF))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Utility functions for SchoolTool calendaring.

These include various date manipulation routines.
"""

import re
import calendar
from datetime import date, datetime, timedelta, time

from pytz import timezone, utc
from zope.testing.cleanup import addCleanUp


def prev_month(date):
    """Calculate the first day of the previous month for a given date.

        >>> prev_month(date(2004, 8, 1))
        datetime.date(2004, 7, 1)
        >>> prev_month(date(2004, 8, 31))
        datetime.date(2004, 7, 1)
        >>> prev_month(date(2004, 12, 15))
        datetime.date(2004, 11, 1)
        >>> prev_month(date(2005, 1, 28))
        datetime.date(2004, 12, 1)

    """
    return (date.replace(day=1) - timedelta(1)).replace(day=1)


def next_month(date):
    """Calculate the first day of the next month for a given date.

        >>> next_month(date(2004, 8, 1))
        datetime.date(2004, 9, 1)
        >>> next_month(date(2004, 8, 31))
        datetime.date(2004, 9, 1)
        >>> next_month(date(2004, 12, 15))
        datetime.date(2005, 1, 1)
        >>> next_month(date(2004, 2, 28))
        datetime.date(2004, 3, 1)
        >>> next_month(date(2004, 2, 29))
        datetime.date(2004, 3, 1)
        >>> next_month(date(2005, 2, 28))
        datetime.date(2005, 3, 1)

    """
    return (date.replace(day=28) + timedelta(7)).replace(day=1)


def week_start(date, first_day_of_week=0):
    """Calculate the first day of the week for a given date.

    Assuming that week starts on Mondays:

        >>> week_start(date(2004, 8, 19))
        datetime.date(2004, 8, 16)
        >>> week_start(date(2004, 8, 15))
        datetime.date(2004, 8, 9)
        >>> week_start(date(2004, 8, 14))
        datetime.date(2004, 8, 9)
        >>> week_start(date(2004, 8, 21))
        datetime.date(2004, 8, 16)
        >>> week_start(date(2004, 8, 22))
        datetime.date(2004, 8, 16)
        >>> week_start(date(2004, 8, 23))
        datetime.date(2004, 8, 23)

    Assuming that week starts on Sundays:

        >>> import calendar
        >>> week_start(date(2004, 8, 19), calendar.SUNDAY)
        datetime.date(2004, 8, 15)
        >>> week_start(date(2004, 8, 15), calendar.SUNDAY)
        datetime.date(2004, 8, 15)
        >>> week_start(date(2004, 8, 14), calendar.SUNDAY)
        datetime.date(2004, 8, 8)
        >>> week_start(date(2004, 8, 21), calendar.SUNDAY)
        datetime.date(2004, 8, 15)
        >>> week_start(date(2004, 8, 22), calendar.SUNDAY)
        datetime.date(2004, 8, 22)
        >>> week_start(date(2004, 8, 23), calendar.SUNDAY)
        datetime.date(2004, 8, 22)

    """
    assert 0 <= first_day_of_week < 7
    delta = date.weekday() - first_day_of_week
    if delta < 0:
        delta += 7
    return date - timedelta(delta)


def weeknum_bounds(year, weeknum):
    """Calculate the inclusive date bounds for a (year, weeknum) tuple.

    Week numbers are as defined in ISO 8601 and returned by
    datetime.date.isocalendar().

        >>> weeknum_bounds(2003, 52)
        (datetime.date(2003, 12, 22), datetime.date(2003, 12, 28))
        >>> weeknum_bounds(2004, 1)
        (datetime.date(2003, 12, 29), datetime.date(2004, 1, 4))
        >>> weeknum_bounds(2004, 2)
        (datetime.date(2004, 1, 5), datetime.date(2004, 1, 11))

    """
    # The first week of a year is at least 4 days long, so January 4th
    # is in the first week.
    firstweek = week_start(date(year, 1, 4), calendar.MONDAY)
    # move forward to the right week number
    weekstart = firstweek + timedelta(weeks=weeknum-1)
    weekend = weekstart + timedelta(days=6)
    return (weekstart, weekend)


def check_weeknum(year, weeknum):
    """Check to see whether a (year, weeknum) tuple refers to a real
    ISO week number.

        >>> check_weeknum(2004, 1)
        True
        >>> check_weeknum(2004, 53)
        True
        >>> check_weeknum(2004, 0)
        False
        >>> check_weeknum(2004, 54)
        False
        >>> check_weeknum(2003, 52)
        True
        >>> check_weeknum(2003, 53)
        False

    """
    weekstart, weekend = weeknum_bounds(year, weeknum)
    isoyear, isoweek, isoday = weekstart.isocalendar()
    return (year, weeknum) == (isoyear, isoweek)


def parse_date(value):
    """Parse a ISO-8601 YYYY-MM-DD date value.

    Examples:

        >>> parse_date('2003-09-01')
        datetime.date(2003, 9, 1)
        >>> parse_date('20030901')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '20030901'
        >>> parse_date('2003-IX-01')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-IX-01'
        >>> parse_date('2003-09-31')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-09-31'
        >>> parse_date('2003-09-30-15-42')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-09-30-15-42'

    """
    try:
        y, m, d = map(int, value.split('-'))
        return date(y, m, d)
    except ValueError:
        raise ValueError("Invalid date: %r" % value)


def parse_datetime(s):
    """Parse a ISO 8601 date/time value.

    Only a small subset of ISO 8601 is accepted:

      YYYY-MM-DD HH:MM:SS
      YYYY-MM-DD HH:MM:SS.ssssss
      YYYY-MM-DDTHH:MM:SS
      YYYY-MM-DDTHH:MM:SS.ssssss

    Returns a datetime.datetime object without a time zone.

    Examples:

        >>> parse_datetime('2003-04-05 11:22:33.456789')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456789)

        >>> parse_datetime('2003-04-05 11:22:33.456')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456000)

        >>> parse_datetime('2003-04-05 11:22:33.45678999')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456789)

        >>> parse_datetime('01/02/03')
        Traceback (most recent call last):
          ...
        ValueError: Bad datetime: 01/02/03

    """
    m = re.match(r"(\d+)-(\d+)-(\d+)[ T](\d+):(\d+):(\d+)([.](\d+))?$", s)
    if not m:
        raise ValueError("Bad datetime: %s" % s)
    ssssss = m.groups()[7]
    if ssssss:
        ssssss = int((ssssss + "00000")[:6])
    else:
        ssssss = 0
    y, m, d, hh, mm, ss = map(int, m.groups()[:6])
    return datetime(y, m, d, hh, mm, ss, ssssss)


def parse_datetimetz(s):
    """Parse a ISO 8601 date/time value in UTC.

    Only a small subset of ISO 8601 is accepted:

      YYYY-MM-DD HH:MM:SS
      YYYY-MM-DD HH:MM:SS.ssssss
      YYYY-MM-DDTHH:MM:SS
      YYYY-MM-DDTHH:MM:SS.ssssss

    optionally followed by the letter Z to indicate UTC time.

    Returns a datetime.datetime object with tzinfo=UTC.

    Examples:

        >>> dt1 = parse_datetimetz('2003-04-05 11:22:33.456789Z')
        >>> dt1.date()
        datetime.date(2003, 4, 5)
        >>> dt1.time()
        datetime.time(11, 22, 33, 456789)
        >>> dt1.tzname()
        'UTC'

        >>> dt2 = parse_datetimetz('2003-04-05 11:22:33.456')
        >>> dt2.date()
        datetime.date(2003, 4, 5)
        >>> dt2.time()
        datetime.time(11, 22, 33, 456000)

        >>> dt3 = parse_datetimetz('2003-04-05 11:22:33.45678999')
        >>> dt3.date()
        datetime.date(2003, 4, 5)
        >>> dt3.time()
        datetime.time(11, 22, 33, 456789)
        >>> dt3.tzname()
        'UTC'

        >>> dt4 = parse_datetimetz('2003-04-05 11:22:33+00:00')
        >>> dt4.date()
        datetime.date(2003, 4, 5)

        >>> dt5 = parse_datetimetz('2003-04-05 11:22:33.45678999-09:00')
        >>> dt5.time()
        datetime.time(11, 22, 33, 456789)

        >>> dt6 = parse_datetimetz('01/02/03')
        Traceback (most recent call last):
          ...
        ValueError: Bad datetime: 01/02/03

    """
    m = re.match(r"(\d+)-(\d+)-(\d+)[ T]"
                 r"(\d+):(\d+):(\d+)([.](\d+))?([-+](\d+):(\d+))?Z?$", s)
    if not m:
        raise ValueError("Bad datetime: %s" % s)
    ssssss = m.groups()[7]
    if ssssss:
        ssssss = int((ssssss + "00000")[:6])
    else:
        ssssss = 0
    y, m, d, hh, mm, ss = map(int, m.groups()[:6])
    return datetime(y, m, d, hh, mm, ss, ssssss, tzinfo=utc)


def parse_time(s):
    """Parse a ISO 8601 time value.

    Only a small subset of ISO 8601 is accepted:

      HH:MM
      HH:MM:SS

    Returns a datetime.time object without a time zone.

    Examples:

        >>> parse_time('11:22:33')
        datetime.time(11, 22, 33)

        >>> parse_time('11:22')
        datetime.time(11, 22)

        >>> parse_time('11:66')
        Traceback (most recent call last):
          ...
        ValueError: minute must be in 0..59

    """
    parts = s.split(":")
    hh, mm = map(int, parts[:2])
    ss = 0
    if len(parts) > 2:
        ss = int(parts[2])
    return time(hh, mm, ss)


def parse_timetz(s, tz=utc):
    """Timezone aware time parser.

    If no timezone preference is given, default to UTC.

        >>> t1 = parse_timetz('11:22:33')
        >>> t1.hour
        11
        >>> t1.minute
        22
        >>> t1.second
        33
        >>> t1.tzname()
        'UTC'

        >>> t2 = parse_timetz('11:22')
        >>> t2.hour
        11
        >>> t2.minute
        22
        >>> t2.second
        0
        >>> t2.tzname()
        'UTC'

        >>> eastern = timezone('US/Eastern')
        >>> teastern = parse_timetz('11:22', tz=eastern)
        >>> teastern.tzname()
        'US/Eastern'

        >>> parse_timetz('11:66')
        Traceback (most recent call last):
          ...
        ValueError: minute must be in 0..59


    """
    parts = s.split(":")
    hh, mm = map(int, parts[:2])
    ss = 0
    if len(parts) > 2:
        ss = int(parts[2])
    return time(hh, mm, ss, tzinfo=tz)


# A hook for unit tests and functional tests:  overrides the "now"
# utcnow() will return its return value if it is not None
_utcnow_hook = None


def utcnow():
    """Return the datetime of now with a UTC timezone.

        >>> tick = utc.localize(datetime.utcnow())
        >>> tack = utcnow()
        >>> tock = utc.localize(datetime.utcnow())
        >>> tick <= tack <= tock
        True

        >>> tack.tzinfo
        <UTC>

    Is hookable for tests by setting the utcnow_hook module global:

        >>> from schooltool.calendar import utils
        >>> utils._utcnow_hook = lambda: datetime(1970, 1, 1, 1, 1, tzinfo=utc)
        >>> utcnow()
        datetime.datetime(1970, 1, 1, 1, 1, tzinfo=<UTC>)

    Clean up:

        >>> utils._utcnow_hook = None
    """
    global _utcnow_hook
    if _utcnow_hook is not None:
        return _utcnow_hook()
    else:
        return utc.localize(datetime.utcnow())


def stub_utcnow(value):
    """Provide a fake value that utcnow will return.

    The fake value can be a datetime:

        >>> stub_utcnow(datetime(1970, 1, 1, 0, 0, tzinfo=utc))
        >>> utcnow()
        datetime.datetime(1970, 1, 1, 0, 0, tzinfo=<UTC>)
        >>> utcnow()
        datetime.datetime(1970, 1, 1, 0, 0, tzinfo=<UTC>)

    Alternatively, it can be a callable:

        >>> def faketime():
        ...    yield datetime(1970, 1, 1, 0, 0, tzinfo=utc)
        ...    yield datetime(1970, 1, 1, 0, 1, tzinfo=utc)
        ...    yield datetime(1970, 1, 1, 0, 2, tzinfo=utc)

        >>> stub_utcnow(faketime().next)
        >>> utcnow()
        datetime.datetime(1970, 1, 1, 0, 0, tzinfo=<UTC>)
        >>> utcnow()
        datetime.datetime(1970, 1, 1, 0, 1, tzinfo=<UTC>)
        >>> utcnow()
        datetime.datetime(1970, 1, 1, 0, 2, tzinfo=<UTC>)

    When we set the stubbed value to None, utcnow starts returning real time:

        >>> stub_utcnow(None)

        >>> tick = utc.localize(datetime.utcnow())
        >>> tack = utcnow()
        >>> tock = utc.localize(datetime.utcnow())
        >>> tick <= tack <= tock
        True

    """
    global _utcnow_hook
    if value is None:
        _utcnow_hook = None
    elif callable(value):
        _utcnow_hook = value
    else:
        _utcnow_hook = lambda: value


addCleanUp(stub_utcnow, (None, ))



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Selenium functional tests for schooltool.app.app
"""
import os
import unittest

from schooltool.app.testing import app_selenium_layer
from schooltool.app.testing import app_selenium_oldskin_layer
from schooltool.testing.selenium import collect_ftests


testdir = os.path.dirname(__file__)
oldskin = ('app.txt',)
stests = [fn for fn in os.listdir(testdir)
          if (fn.endswith('.txt') and
              not fn.startswith('.') and
              not fn in oldskin)]


def test_suite():
    return unittest.TestSuite([
        collect_ftests(layer=app_selenium_layer, filenames=stests),
        collect_ftests(layer=app_selenium_oldskin_layer, filenames=oldskin),
        ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from schooltool.testing.util import format_table
from schooltool.testing.analyze import queryHTML
from schooltool.testing.functional import ZCMLLayer
from schooltool.testing.selenium import SeleniumLayer

here = os.path.dirname(__file__)

app_functional_layer = ZCMLLayer(os.path.join(here, 'ftesting.zcml'),
                                 __name__,
                                 'app_functional_layer')

app_selenium_layer = SeleniumLayer(os.path.join(here, 'stesting.zcml'),
                                   __name__+'_selenium',
                                   'app_functional_layer')

app_selenium_oldskin_layer = SeleniumLayer(os.path.join(here,
                                           'stesting-oldskin.zcml'),
                                           __name__+'_oldskin',
                                           'app_functional_layer')


def format_weekly_calendar(contents):
    table = []
    for n, row in enumerate(queryHTML('//table[@id="calendar-view-week"]//tr', contents)):
        if n == 0:
            header = []
            for cell in queryHTML('//tr//th/a/text()', str(row)):
                header.append(cell)
            table.append(header)
        else:
            block = []
            for cell in queryHTML('//tr//td', str(row)):
                events = []
                for event in queryHTML('//td//a//span/text()', str(cell)):
                    events.append(event)
                block.append(", ".join(events))
            table.append(block)
    return format_table(table, header_rows=1)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Functional Testing Utilities for schoolyears
"""
import os

from schooltool.testing.functional import ZCMLLayer

dir = os.path.abspath(os.path.dirname(__file__))
filename = os.path.join(dir, 'ftesting.zcml')

course_functional_layer = ZCMLLayer(filename,
                                    __name__,
                                    'course_functional_layer')
